/*
 aiwolf.hpp
 Katsuki Ohto
 */

#ifndef AIWOLF_AIWOLF_HPP_
#define AIWOLF_AIWOLF_HPP_

#include <cstring>
#include <unistd.h>
#include <sys/time.h>
#include <ctime>

#include <cmath>
#include <cstdio>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <cassert>
#include <thread>
#include <mutex>
#include <atomic>
#include <array>
#include <vector>
#include <queue>
#include <map>
#include <random>
#include <algorithm>
#include <string>
#include <bitset>

#ifdef _WIN32

#include <winsock2.h>
#include <ws2tcpip.h>

#else

#include <arpa/inet.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>

#endif

// 設定
#include "settings.h"

// 基本的定義、ユーティリティ
// 使いそうなものは全て読み込んでおく

// 乱数
#include "../../common/util/xorShift.hpp"

// 確率分布
#include "../../common/util/random.hpp"

// ビット演算
#include "../../common/util/bitOperation.hpp"

// 包括ユーティリティ
#include "../../common/util/container.hpp"
#include "../../common/util/arrays.h"

// クラスユーティリティ
#include "../../common/util/bitSet.hpp"
#include "../../common/util/bitArray.hpp"

// 応用ユーティリティ
#include "../../common/util/bitPartition.hpp"

// 解析用
#include "../../common/util/math.hpp"
#include "../../common/util/pd.hpp"

// 探索
#include "../../common/util/search.hpp"

// ハッシュ
#include "../../common/hash/HashFunc.hpp"
#include "../../common/hash/HashBook.hpp"

// 多腕バンディット問題
#include "../../common/util/multiArmedBandit.hpp"

// softmax方策
#include "../../common/util/softmaxPolicy.hpp"

// 並列化
#include "../../common/util/lock.hpp"

namespace AIWolf{
    
    /**************************ルール**************************/
    
    constexpr int N_TALK_TURNS = 10;
    constexpr int N_WHISPER_TURNS = 10;
    constexpr int N_MAX_PLAYERS = 16; // 実際の人数は試合ごとに違う
    constexpr int N_MAX_DAYS = 16;

    struct MatchRule{ // 対戦毎に通知するルール
        int players; // プレーヤー人数
        int turns; // 会話、狼同士の会話ターン数
        BitArray64<8, 8> rolePlayers; // 役職毎の人数
        
        void clear()noexcept{
            players = 0; turns = 0;
            rolePlayers.clear();
        }
        
        std::string toString()const{
            std::ostringstream oss;
            oss << "players = " << players << " turns = " << turns << " rolePlayers = " << rolePlayers << endl;
            return oss.str();
        }
        
        constexpr MatchRule()
        : players(), turns(), rolePlayers(){}
        
        constexpr MatchRule(const MatchRule& mr)
        : players(mr.players), turns(mr.turns), rolePlayers(mr.rolePlayers){}
        
        constexpr MatchRule(int p, int t, BitArray64<8, 8> rp)
        : players(p), turns(t), rolePlayers(rp){}
    };
    constexpr int NONE_DAY = -1;
    constexpr int examDay(int d)noexcept{
        return (0 <= d && d < N_MAX_DAYS);
    }
    
    constexpr int DAY_FIRST = 0;
    
    int toDayNum(const std::string& str){
        if(str.size() < 4){ return NONE_DAY; }
        int d = atoi(str.substr(3, str.size()).c_str());
        if(!examDay(d)){ return NONE_DAY; }
        return d;
    }
    
    constexpr int DAY_ATTACK_FIRST = 1;
    constexpr int DAY_EXECUTE_FIRST = 1;
    constexpr int DAY_DIVINE_FIRST = 0;
    
    /**************************役職**************************/
    
    namespace Role{
        enum{
            VILLAGER  = 0, // 村人
            SEER      = 1, // 占い師
            MEDIUM    = 2, // 霊媒師
            BODYGUARD = 3, // 狩人
            POSSESSED = 4, // 狂人
            WEREWOLF  = 5, // 人狼
        };
    }
    constexpr int N_ROLES = 6;
    constexpr int NONE_ROLE = N_ROLES;
    
    constexpr int N_MAX_ROLE_PLAYERS[] = {8, 1, 1, 1, 1, 3,};
    
    constexpr int maxPlayerNum(int r){ return N_MAX_ROLE_PLAYERS[r]; }
    
    const std::string roleString[] = {
        "VILLAGER", "SEER", "MEDIUM", "BODYGUARD",
        "POSSESSED", "WEREWOLF", "none",
    };
    const char roleChar[] = {
        'V', 'S', 'M', 'B', 'P', 'W', '.',
    };
    
    bool examRole(int r){
        return (0 <= r && r < N_ROLES + 1);
    }
    bool examValidRole(int r){
        return (0 <= r && r < N_ROLES);
    }
    
    std::string toRoleString(int r){
        if(!examRole(r)){ r = NONE_ROLE; }
        return roleString[r];
    }
    
    // 所属チーム
    constexpr static int getTeam(int r)noexcept{
        return (r >= Role::POSSESSED);
    }
    // 種族は現在2通りしかないので人狼の時だけ Species::WEREWOLF(=1) を返す
    constexpr static int getSpecies(int r)noexcept{
        return (r == Role::WEREWOLF);
    }
    
    template<class callback_t>
    void iterateValidRole(const callback_t& callback){
        callback(Role::VILLAGER);
        callback(Role::WEREWOLF);
        callback(Role::SEER);
        callback(Role::MEDIUM);
        callback(Role::POSSESSED);
        callback(Role::BODYGUARD);
    }
    template<class callback_t>
    int searchValidRole(const callback_t& callback){
        if(callback(Role::VILLAGER)){ return Role::VILLAGER; }
        if(callback(Role::WEREWOLF)){ return Role::WEREWOLF; }
        if(callback(Role::SEER)){ return Role::SEER; }
        if(callback(Role::MEDIUM)){ return Role::MEDIUM; }
        if(callback(Role::POSSESSED)){ return Role::POSSESSED; }
        if(callback(Role::BODYGUARD)){ return Role::BODYGUARD; }
        return NONE_ROLE;
    }
    
    int toRoleNum(const std::string& str){
        return searchValidRole([&str](int r){
            return (str == toRoleString(r));
        });
    }
    
    /**************************種族**************************/
    
    namespace Species{
        enum{
            HUMAN     = 0, // 人間
            WEREWOLF  = 1, // 人狼
        };
    }
    constexpr int N_SPECIES = 2;
    constexpr int NONE_SPECIES = 2;
    
    // 種族最大人数
    constexpr int N_MAX_HUMAN_SPECIES = maxPlayerNum(Role::VILLAGER) + maxPlayerNum(Role::SEER)
    + maxPlayerNum(Role::MEDIUM) + maxPlayerNum(Role::BODYGUARD) + maxPlayerNum(Role::POSSESSED);
    
    constexpr int N_MAX_WEREWOLF_SPECIES = maxPlayerNum(Role::WEREWOLF);
    
    const std::string speciesString[] = {
        "HUMAN", "WEREWOLF",
    };
    constexpr char speciesChar[] = {
       'O', 'X',
    };
    
    constexpr bool examSpecies(int s)noexcept{
        return !(s & (~1));
    }
    
    std::string toSpeciesString(int s){
        return speciesString[s & 1];
    }
    int toSpeciesNum(const std::string& str){
        if(str == speciesString[Species::HUMAN]){
            return Species::HUMAN;
        }else if(str == speciesString[Species::WEREWOLF]){
            return Species::WEREWOLF;
        }
        return NONE_SPECIES;
    }
    
    /**************************チーム**************************/
    
    namespace Team{
        enum{
            VILLAGER = 0, // 村人陣営
            WEREWOLF = 1, // 人狼陣営
        };
    };
    constexpr int N_TEAMS = 2;
    constexpr int NONE_TEAM = 2;
    
    // チーム最大人数
    constexpr int N_MAX_VILLAGER_TEAM = maxPlayerNum(Role::VILLAGER) + maxPlayerNum(Role::SEER)
    + maxPlayerNum(Role::MEDIUM) + maxPlayerNum(Role::BODYGUARD);
    
    constexpr int N_MAX_WEREWOLF_TEAM = maxPlayerNum(Role::WEREWOLF) + maxPlayerNum(Role::POSSESSED);
    
    const std::string teamString[] = {
        "VILLAGER", "WEREWOLF",
    };
    constexpr char teamChar[] = {
        '+', '-',
    };
    
    bool examTeam(int t){
        return !(t & (~1));
    }
    
    std::string toTeamString(int t){
        return teamString[t & 1];
    }
    int toTeamNum(const std::string& str){
        if(str == teamString[Team::VILLAGER]){
            return Team::VILLAGER;
        }else if(str == teamString[Team::WEREWOLF]){
            return Team::WEREWOLF;
        }
        return NONE_TEAM;
    }
    
    /**************************エージェント**************************/
    
    constexpr int toOfficialAgentNumber(int n)noexcept{
        return n + 1;
    }
    constexpr int toMyAgentNumber(int n)noexcept{
        return n - 1;
    }
    
    constexpr bool examAgentNum(int n)noexcept{
        return (0 <= n && n < N_MAX_PLAYERS);
    }
    
    std::string toAgentString(int i){
        i = toOfficialAgentNumber(i);
        std::ostringstream oss;
        oss << "Agent[";
        if(i < 10){
            oss << "0";
        }
        oss << i << "]";
        return oss.str();
    }
    std::string toAgentJsonString(int i){
        i = toOfficialAgentNumber(i);
        std::ostringstream oss;
        oss << "{\"agentIdx\":";
        if(i < 10){
            oss << "0";
        }
        oss << i << "}";
        return oss.str();
    }
    constexpr int NONE_PLAYER = N_MAX_PLAYERS;
    
    int toAgentNum(const std::string& str){
        if(str.size() < 9){ return NONE_PLAYER; }
        //char i1 = str[6], i0 = str[7];
        //int p = toMyAgentNumber(int(i1 - '0') * 10 + int(i0 - '0'));
        int p = toMyAgentNumber(atoi(str.substr(6, 8).c_str()));
        //cerr << "p = " << p << endl;
        if(!examAgentNum(p)){ return NONE_PLAYER; }
        return p;
    }
    
    /**************************エージェントの状態**************************/
    
    enum PlayerState{
        ALIVE = 0,
        EXECUTED,
        ATTACKED,
    };
    
    /**************************発話タイプ**************************/
    
    // 共通
    enum StatementType{
        OVER = 0, // 発言終了
        SKIP = 1, // 無発言
        ESTIMATE, // PはRだと思う
        COMINGOUT, // 私はRだ
        DIVINED, // Pを占った結果Rだった
        INQUESTED, // Pの霊媒結果はRだった
        GUARDED, // Pを護衛した
        VOTE, // Pに投票する
        AGREE, // Iに同意する
        DISAGREE, // Iに反対する
    };
    enum StatementCategory{
        TALK = 0,
        WHISPER,
    };
    
    constexpr int N_STATEMENTTYPES = 10;
    
    // 表の会話
    constexpr int N_TALKTYPES = N_STATEMENTTYPES + 0;
    
    constexpr bool examTalkType(int t)noexcept{
        return (0 <= t && t < N_TALKTYPES);
    }
    
    // 狼の会話
    enum WhisperType{
        ATTACK = N_STATEMENTTYPES, // Pを襲撃する
    };
    constexpr int N_WHISPERTYPES = N_STATEMENTTYPES + 1;
    
    constexpr bool examWhisperType(int t)noexcept{
        return (0 <= t && t < N_WHISPERTYPES);
    }
    
    constexpr bool examStatementType(int t)noexcept{
        return (0 <= t && t < max(N_TALKTYPES, N_WHISPERTYPES));
    }
    
    const std::string statementString[] = {
        "Over", "Skip", "ESTIMATE", "COMINGOUT",
        "DIVINED", "INQUESTED", "GUARDED", "VOTE",
        "AGREE", "DISAGREE", "ATTACK",
    };
    const std::string statementCategoryString[] = {
        "TALK", "WHISPER",
    };
    constexpr char statementChar[] = { // agree -> y (yes), disagree-> n (no)
        'o', 's', 'e', 'c', 'd', 'i', 'g', 'v', 'y', 'n', 'a',
    };
    constexpr char statementCategoryChar[] = {
        '/', '\\',
    };
    
    std::string toCategoryString(int c){
        return statementCategoryString[c & 1];
    }
    
    constexpr int NONE_CATEGORY = -1;
    constexpr int ILLEGAL_STATEMENT = -1;
    
    int toCategoryNum(const std::string& str){
        if(str == statementCategoryString[TALK]){
            return TALK;
        }else if(str == statementCategoryString[WHISPER]){
            return WHISPER;
        }
        return NONE_CATEGORY;
    }
    
    template<class callback_t>
    int searchStatementType(const callback_t& callback){
        for(int t = OVER; t <= ATTACK; ++t){
            if(callback(t)){ return t; }
        }
        return ILLEGAL_STATEMENT;
    }
    
    /**************************会話のインデックス**************************/
    
    // 使用変更
    /*constexpr int getTalkDay(int i)noexcept{
        return i / (N_MAX_DAYS * N_TALK_TURNS * N_MAX_PLAYERS);
    }
    constexpr int getTalkIndexInDay(int i)noexcept{
        return i % (N_MAX_DAYS * N_TALK_TURNS * N_MAX_PLAYERS);
    }*/
    
    /**************************会話**************************/
    
    struct Statement : public BitArray32<4, 8>{
        // 28ビットで発話内容を表現
        // 4ビットは他の情報(プレーヤー番号)などに使う
        
        constexpr Statement() : BitArray32<4, 8>(){}
        constexpr Statement(uint32 i) : BitArray32<4, 8>(i){}
        
        constexpr static uint32 CONTENT_MASK = (1U << 28) - 1U;
        
        constexpr bool isOver()const noexcept{
            return ((data() & CONTENT_MASK)  == StatementType::OVER);
        }
        constexpr bool isSkip()const noexcept{
            return ((data() & CONTENT_MASK) == StatementType::SKIP);
        }
        
        constexpr bool isPlainOver()const noexcept{ return (data() == StatementType::OVER); }
        constexpr bool isPlainSkip()const noexcept{ return (data() == StatementType::SKIP); }
        
        constexpr auto any()const noexcept{ // skip, overでない何らかの発言である
            return data() & (~1);
        }
        constexpr auto type()const noexcept{ // 会話タイプ
            return (*this)[0];
        }
        constexpr auto dst()const noexcept{ // 言及するプレーヤー
            return (*this)[1];
        }
        constexpr auto role()const noexcept{ // 言及する役職(or種族)
            return (*this)[2];
        }
        constexpr auto category()const noexcept{ // 言及する会話カテゴリ
            return (*this)[5];
        }
        constexpr auto day()const noexcept{ // 言及する会話日付
            return (*this)[6];
        }
        auto index()const noexcept{ // 言及する会話インデックス
            return at(1, 4);
        }
        constexpr auto player()const noexcept{ // プレーヤー番号
            return (*this)[7];
        }
        
        void setType(int t)noexcept{ set(0, t); }
        void setDst(int p)noexcept{ set(1, p); }
        void setRole(int r)noexcept{ set(2, r); }
        void setCategory(int i)noexcept{ set(5, i); }
        void setDay(int i)noexcept{ set(6, i); }
        void setIndex(int i)noexcept{ set(1, i); }
        void setPlayer(int p)noexcept{ set(7, p); }
        
        void setVO(int v, int o)noexcept{ setType(v); setDst(o); }
        //void setVC(int v, int c)noexcept{ setType(v); setRole(c); }
        void setVOC(int v, int o, int c)noexcept{ setType(v); setDst(o); setRole(c); }
        void setVDCI(int v, int d, int c, int i)noexcept{
            setType(v); setDay(d); setCategory(c); setIndex(i);
        }
        template<int OUT = 0>
        bool exam()const{
            if(!examStatementType(type())){
                if(OUT){ cerr << "Statement::exam() : illegal statement type! " << type() << endl; }
                return false;
            }
            return true;
        }
    };
    
    std::ostream& operator<<(std::ostream& os, const Statement& stm){
        // 短い出力
        if(!stm.exam()){ os << '*'; } // おかしな値
        int t = stm.type();
        os << statementChar[t];
        switch(t){
            case VOTE: case GUARDED: case ATTACK:{
                os << stm.dst();
            }break;
            case COMINGOUT: case ESTIMATE:{
                os << stm.dst() << roleChar[stm.role()];
            }break;
            case DIVINED: case INQUESTED:{
                os << stm.dst() << speciesChar[stm.role()];
            }break;
            case AGREE: case DISAGREE:{
                os << stm.day() << statementCategoryChar[stm.category()];
                os << stm.index();
            }break;
            default: break;
        }
        return os;
    }
    
    template<class statement_t>
    std::string toStatementString(const statement_t& stm){
        // 通信用
        if(!stm.exam()){ return statementString[OVER]; } // おかしな値
        int t = stm.type();
        std::stringstream oss;
        oss << statementString[t];
        switch(t){
            case VOTE: case GUARDED: case ATTACK:{
                oss << " " << toAgentString(stm.dst());
            }break;
            case COMINGOUT: case ESTIMATE:{
                oss << " " + toAgentString(stm.dst()) << " " << toRoleString(stm.role());
            }break;
            case DIVINED: case INQUESTED:{
                oss << " " + toAgentString(stm.dst()) << " " << toSpeciesString(stm.role());
            }break;
            case AGREE: case DISAGREE:{
                oss << " " << toCategoryString(stm.category());
                oss << " " << "day" << stm.day() << " ID:" << stm.index();
            }break;
            default: break;
        }
        return oss.str();
    }
    
    int toStatementType(const std::string& str){
        return searchStatementType([&str](int t){
            return (str == statementString[t]);
        });
    }
    
    int toIndex(const std::string& str){
        if(str.size() < 4){ return -1; }
        int d = atoi(str.substr(3, str.size()).c_str());
        return d;
    }
    
    Statement toStatementData(const std::string& str, const int category){
        Statement stm;
        std::vector<std::string> v = split(str, ' ');
        if(v.size() <= 0){ return ILLEGAL_STATEMENT; }
        int type = toStatementType(v[0]);
        //if(type == INQUESTED){ cerr << "i" << endl; }
        if(type == ILLEGAL_STATEMENT){ return ILLEGAL_STATEMENT; }
        stm.setType(type);
        switch(type){
            case VOTE: case GUARDED: case ATTACK:{
                if(category == TALK && type == ATTACK){ return ILLEGAL_STATEMENT; }
                if(v.size() < 2){ return ILLEGAL_STATEMENT; }
                int p = toAgentNum(v[1]); if(p == NONE_PLAYER){ return ILLEGAL_STATEMENT; }
                stm.setDst(p);
            }break;
            case COMINGOUT: case ESTIMATE:{
                if(v.size() < 3){ return ILLEGAL_STATEMENT; }
                int p = toAgentNum(v[1]); if(p == NONE_PLAYER){ return ILLEGAL_STATEMENT; }
                stm.setDst(p);
                int r = toRoleNum(v[2]); if(r == NONE_ROLE){ return ILLEGAL_STATEMENT; }
                stm.setRole(r);
            }break;
            case DIVINED: case INQUESTED:{
                if(v.size() < 3){ return ILLEGAL_STATEMENT; }
                int p = toAgentNum(v[1]); if(p == NONE_PLAYER){ return ILLEGAL_STATEMENT; }
                stm.setDst(p);
                int s = toSpeciesNum(v[2]); if(s == NONE_SPECIES){ return ILLEGAL_STATEMENT; }
                stm.setRole(s);
            }break;
            case AGREE: case DISAGREE:{
                if(v.size() < 4){ return ILLEGAL_STATEMENT; }
                int c = toCategoryNum(v[1]); if(c == NONE_CATEGORY){ return ILLEGAL_STATEMENT; }
                stm.setCategory(c);
                int d = toDayNum(v[2]); if(d == NONE_DAY){ return ILLEGAL_STATEMENT; }
                stm.setDay(d);
                int i = toIndex(v[3]); if(i < 0){ return ILLEGAL_STATEMENT; }
                stm.setIndex(i);
            }break;
            default: break; // overやskip
        }
        return stm;
    }
    
    struct Talk : public Statement{
        constexpr Talk() : Statement(){}
        constexpr Talk(uint32 i) : Statement(i){}
        constexpr Talk(const Statement& stm) : Statement(stm){}
        
        template<int OUT = 0>
        bool exam()const{
            if(!examTalkType(type())){
                if(OUT){ cerr << "talk::exam() : illegal statement type! " << type() << endl; }
                return false;
            }
            return true;
        }
    };
    
    struct Whisper : public Statement{
        constexpr Whisper() : Statement(){}
        constexpr Whisper(uint32 i) : Statement(i){}
        constexpr Whisper(const Statement& stm) : Statement(stm){}
        
        template<int OUT = 0>
        bool exam()const{
            if(!examWhisperType(type())){
                if(OUT){ cerr << "Whisper::exam() : illegal statement type! " << type() << endl; }
                return false;
            }
            return true;
        }
    };
    
    template<class statement_t, class info_t>
    int genAllStatements(statement_t *const pst0, const info_t& info, int category){
        // 生成しうる全ての発話を生成(comingoutで他人を指定したりなどおかしなものを含む)
        statement_t *pst = pst0;
        pst->setType(OVER); ++pst;
        pst->setType(SKIP); ++pst;
        
        const int P = info.rule().players;
        // SVO型
        for(int p = 0; p < P; ++p){
            pst->setVO(VOTE, p); ++pst;
            pst->setVO(GUARDED, p); ++pst;
            if(category == WHISPER){
                pst->setVO(ATTACK, p); ++pst;
            }
        }
        // SVOC型
        for(int p = 0; p < P; ++p){
            iterateValidRole([p, &pst, &info](int r){
                pst->setVOC(COMINGOUT, p, r); ++pst;
                pst->setVOC(ESTIMATE, p, r); ++pst;
            });
            for(int s = 0; s < N_SPECIES; ++s){
                pst->setVOC(DIVINED, p, s); ++pst;
                pst->setVOC(INQUESTED, p, s); ++pst;
            }
        }

        // agree, disagree
        for(int d = 0; d <= info.day(); ++d){
            for(int i = 0; i < info.NTalks(d); ++i){
                pst->setVDCI(AGREE, d, TALK, i); ++pst;
                pst->setVDCI(DISAGREE, d, TALK, i); ++pst;
            }
            if(category == WHISPER){
                for(int i = 0; i < info.NWhispers(d); ++i){
                    pst->setVDCI(AGREE, d, WHISPER, i); ++pst;
                    pst->setVDCI(DISAGREE, d, WHISPER, i); ++pst;
                }
            }
        }
        // 合法性チェック
        for(const statement_t *p = pst0; p != pst; ++p){
            //cerr << (*p) << endl;
            ASSERT(p->template exam<1>(), cerr << (p - pst0) << endl;);
        }
        return pst - pst0;
    }
    
    /**************************盤面**************************/
    
    struct Info{ // 共通の情報
        
        MatchRule rule_;
        
        // <盤の一時状態>
        int game_, day_;
        int talkIndex_;
        int whisperTurn_, whisperIndex_;
        
        // <プレーヤー固定状態>
        BitArray64<4, 16> playerRole_; // プレーヤー -> 役職
        std::array<BitSet16, N_ROLES> rolePlayers_; // 役職->プレーヤー
        
        // <プレーヤー日別状態>
        BitSet16 alive_[N_MAX_DAYS]; // aliveかどうかのbit
        BitSet16 roleAlive_[N_MAX_DAYS][N_ROLES]; // 各役職でaliveかどうかのbit
        
        // <プレーヤー更新型状態>
        BitArray64<4, 16> deathDay_; // 死亡日
        BitArray64<4, 16> causeOfDeath_; // 死因

        // <行動系の情報>
        BitArray64<4, N_MAX_DAYS> attack_; // 襲撃
        BitArray64<4, N_MAX_DAYS> divine_; // 占い
        BitArray64<4, N_MAX_DAYS> guard_; // 護衛
        
        // <行動結果の情報>
        BitArray64<4, N_MAX_DAYS> executed_; // 処刑
        BitArray64<4, N_MAX_DAYS> attacked_; // 襲撃
        BitArray64<4, N_MAX_DAYS> divined_; // 占い
        BitArray64<4, N_MAX_DAYS> inquested_; // 霊媒
        
        // <行動系のフラグ>
        BitArray64<16, 4> actionFlag; // 0 divine 1 guard 2 attack 3 attacked
        
        // <投票系の情報>
        std::array<BitArray64<4, 16>, N_MAX_DAYS> vote_; // 処刑投票
        std::array<BitArray64<4, N_MAX_PLAYERS>, N_MAX_DAYS> attackVote_; // 襲撃投票
        
        // <会話系の情報>
        // 会話ログ
        Talk talk_[N_MAX_DAYS][N_TALK_TURNS * N_MAX_PLAYERS];
        int NTalks_[N_MAX_DAYS];
        // 人狼内会話ログ
        Whisper whisper_[N_MAX_DAYS][N_TALK_TURNS * N_WHISPER_TURNS * maxPlayerNum(Role::WEREWOLF)];
        int NWhispers_[N_MAX_DAYS];
        int TurnWhisperFirstIndex_[N_MAX_DAYS][N_TALK_TURNS + 1];
        
        // カミングアウト情報
        //BitArray64<4, N_MAX_PLAYERS> latestComingOutRole;
        
        // ルール
        const MatchRule& rule()const noexcept{ return rule_; }
        
        // <状態取得>
        int game()const noexcept{ return game_; }
        int day()const noexcept{ return day_; }
        
        int talkTurn()const noexcept{ return talkIndex_ / rule().players; }
        int talkIndex()const noexcept{ return talkIndex_; }
        
        int whisperTurn()const noexcept{ return whisperTurn_; }
        int whisperIndex()const noexcept{ return whisperIndex_; }
        
        auto isAlive(int d, int p)const{ return alive_[d].test(p); }
        BitSet16 alive(int d)const{ return alive_[d]; }
        int NAlive(int d)const{ return alive_[d].count(); }
        
        auto roleAlive(int d, int r)const{ return roleAlive_[d][r];}
        int NRoleAlive(int d, int r)const{ return roleAlive_[d][r].count(); }
        
        BitSet16 teamAlive(int d, int t)const{
            BitSet16 w = roleAlive(d, Role::WEREWOLF) | roleAlive(d, Role::POSSESSED);
            return (t == Team::VILLAGER) ? BitSet16(alive(d) & (~w)) : w;
        }
        int NTeamAlive(int d, int t)const{ return teamAlive(d, t).count(); }
        
        BitSet16 speciesAlive(int d, int s)const{
            BitSet16 w = roleAlive(d, Role::WEREWOLF);
            return (s == Species::HUMAN) ? BitSet16(alive(d) & (~w)) : w;
        }
        int NSpeciesAlive(int d, int s)const{
            return (s == Species::HUMAN) ? (NAlive(d) - NRoleAlive(d, Role::WEREWOLF)) : NRoleAlive(d, Role::WEREWOLF);
        }
        auto isAliveAfterExecution(int d, int p)const{ // 処刑終了時点で生きているか
            return isAlive(d, p) && (executed(d) != p);
        }
        
        int playerRole(int p)const{ return playerRole_[p]; }
        auto playerRoles()const noexcept{ return playerRole_; }
        auto rolePlayers(int r)const{ return rolePlayers_[r]; }
        
        int deathDay(int p)const{ return deathDay_[p]; }
        int causeOfDeath(int p)const{ return causeOfDeath_[p]; }
        
        // <状態判定>
        bool isAliveOther(int d, int p, int t)const{ return isAlive(d, t) && (p != t); }
        int isGameOver(int d)const{ // 勝敗が決したか
            if(NSpeciesAlive(d, Species::WEREWOLF) == 0){ return Team::VILLAGER; } //村人勝ち
            if(NSpeciesAlive(d, Species::HUMAN) <= NSpeciesAlive(d, Species::WEREWOLF)){
                return Team::WEREWOLF; // 人狼勝ち
            }
            return NONE_TEAM; // まだ試合続行
        }
        
        // <行動>
        int attack(int d)const{ return attack_[d]; }
        int divine(int d)const{ return divine_[d]; }
        int guard(int d)const{ return guard_[d]; }
        
        bool anyAttack(int d)const{ return actionFlag.at_flag(2, d); }
        bool anyDivine(int d, int p)const{ return actionFlag.at_flag(0, d); }
        bool anyGuard(int d, int p)const{ return actionFlag.at_flag(1, d); }
        
        // 行動集合
        auto attackAll()const noexcept{ return attack_; }
        auto divineAll()const noexcept{ return divine_; }
        auto guardAll()const noexcept{ return guard_; }
        
        // <行動結果>
        int executed(int d)const{ return executed_[d]; }
        int attacked(int d)const{ return attacked_[d]; }
        int divined(int d)const{ return divined_[d]; }
        int inquested(int d)const{ return inquested_[d]; }
        
        constexpr static bool anyExecuted(int d)noexcept{ return (d >= DAY_EXECUTE_FIRST); }
        bool anyAttacked(int d)const{ return actionFlag.at_flag(3, d); }

        // 行動結果集合
        auto executedAll()const noexcept{ return executed_; }
        auto attackedAll()const noexcept{ return attacked_; }
        auto divinedAll()const noexcept{ return divined_; }
        auto inquestedAll()const noexcept{ return inquested_; }
        
        // <投票系>
        auto vote(int d, int p)const{ return vote_[d][p]; }
        auto attackVote(int d, int p)const{ return attackVote_[d][p]; }
        auto voteDay(int d)const{ return vote_[d]; }
        auto attackVoteDay(int d)const{ return attackVote_[d]; }
        
        // 投票判定
        bool judgeAttackAbstaination(int d)const noexcept{ // 人狼の全員が襲撃を放棄したか調べる
            return (search(rolePlayers(Role::WEREWOLF), [d, this](int p)->bool{
                return (this->attackVote(d, p) != p);
            }) == -1);
        }
        // 合法性判定
        // 棄権はとりあえずしないので非合法とする
        bool isValidAttackVote(int d, int p, int t)const{
            return isAlive(d, t) && (playerRole(t) != Role::WEREWOLF);
        }
        bool isValidVote(int d, int p, int t)const{
            return isAliveOther(d, p, t);
        }
        bool isValidGuard(int d, int p, int t)const{
            return isAliveOther(d, p, t);
        }
        bool isValidDivine(int d, int p, int t)const{
            return isAlive(d, t); // 占いは自分自身も合法
        }
        bool isValidAttack(int d, int t)const{
            return isAlive(d, t) && (playerRole(t) != Role::WEREWOLF);
        }
        template<class act_t>
        bool isValidAction(int d, int p, const act_t& act)const{
            // 行動権利がない場合の場所の値は不問
            if(d >= DAY_EXECUTE_FIRST){
                if(!isValidVote(d, p, act[0])){ return false; }
            }
            int r = playerRole(p);
            if(r == Role::SEER){
                if(d >= DAY_DIVINE_FIRST){
                    if(!isValidDivine(d, p, act[1])){ return false; }
                }
            }else if(r == Role::BODYGUARD){
                if(d >= DAY_ATTACK_FIRST){
                    if(!isValidGuard(d, p, act[1])){ return false; }
                }
            }else if(r == Role::WEREWOLF){
                if(d >= DAY_ATTACK_FIRST){
                    if(!isValidAttackVote(d, p, act[1])){ return false; }
                }
            }
            return true;
        }
        bool isValidTalk(int d, int p, Talk t)const{
            if(!t.exam()){ return false; }
            return true;
        }
        bool isValidWhisper(int d, int p, Whisper w)const{
            if(!w.exam()){ return false; }
            return true;
        }

        // <会話系>
        Talk talk(int d, int t, int i)const{ return talk_[d][t * NAlive(d) + i]; }
        Talk talk(int d, int i)const{ return talk_[d][i]; }
        int NTalks(int d)const{ return NTalks_[d]; }
        
        Whisper whisper(int d, int i)const{ return whisper_[d][i]; }
        int NWhispers(int d)const{ return NWhispers_[d]; }
        int TurnWhisperFirstIndex(int d, int t)const{ return TurnWhisperFirstIndex_[d][t]; }
        int NTurnWisper(int d, int t)const{ return TurnWhisperFirstIndex_[d][t + 1] - TurnWhisperFirstIndex_[d][t]; }
        
        template<int YN>
        std::pair<Statement, int> searchAgreeChain(int d, int c, int i)const{
            // 同意や反対宣言に対して、元の言明を再帰的に探索し、元の言明と賛成反対を返す
            const Statement stm = (c == TALK) ? Statement(talk(d, i)) : Statement(whisper(d, i));
            if(stm.type() == StatementType::AGREE){
                return searchAgreeChain<YN>(stm.day(), stm.category(), stm.index());
            }else if(stm.type() == StatementType::DISAGREE){
                return searchAgreeChain<1 - YN>(stm.day(), stm.category(), stm.index());
            }else{
                return std::make_pair(stm, YN);
            }
        }
        
        // <状態設定>
        void setAlive(int d, int p){ // ログからの設定用
            alive_[d].set(p);
            if(playerRole(p) != NONE_ROLE){
                roleAlive_[d][playerRole(p)].set(p);
            }
        }
        void setgame(int g)noexcept{ game_ = g; }
        void setDay(int d)noexcept{ day_ = d; }
        void setRole(int p, int r){
            playerRole_.assign(p, r);
            rolePlayers_[r].set(p);
        }
        
        // <行動報告>
        void setAttack(int d, int t){ attack_.assign(d, t); actionFlag.set_flag(2, d); }
        void setAbandonAttack(int d){}
        void setDivine(int d, int p, int t){ divine_.assign(d, t); actionFlag.set_flag(0, d); }
        void setAbandonDivine(int d, int p){ divined_.assign(d, NONE_SPECIES); }
        void setGuard(int d, int p, int t){ guard_.assign(d, t); actionFlag.set_flag(1, d); }
        void setAbandonGuard(int d, int p){}
        
        // <行動結果報告>
        void setExecuted(int d, int p){
            executed_.assign(d, p);
            deathDay_.assign(p, d);
            causeOfDeath_.assign(p, EXECUTED);
        }
        void setAttacked(int d, int p){
            attacked_.assign(d, p);
            deathDay_.assign(p, d);
            causeOfDeath_.assign(p, ATTACKED);
            actionFlag.set_flag(3, d);
        }
        void setDivined(int d, int p, int t, int r){ divined_.assign(d, r); }
        void setInquested(int d, int p, int t, int r){ inquested_.assign(d, r); }
        
        void setNoAttacked(int d){}
        
        // <投票報告>
        void setVote(int d, int p, int t){ vote_[d].assign(p, t); }
        void setAbstainVote(int d, int p){ // 処刑投票棄権
            vote_[d].assign(p, p); // 自分自身を投票することで棄権を表す
        }
        void setAttackVote(int d, int p, int t){ attackVote_[d].assign(p, t); }
        void setAbstainAttackVote(int d, int p){ // 襲撃投票棄権
            attackVote_[d].assign(p, p); // 自分自身を襲撃することで棄権を表す
        }
        
        // <会話報告>
        void setWhisperIndex(int d, int i)noexcept{
            whisperIndex_ = i;
            NWhispers_[d] = i; // その日ここまでの会話数
        }
        void setPlayerTalk(int d, int i, int p, Talk t){
            t.setPlayer(p);
            talk_[d][i] = t;
        }
        void setPlayerWhisper(int d, int i, int p, Whisper w){
            w.setPlayer(p);
            whisper_[d][i] = w;
        }
        void setNTalks(int d, int n){ NTalks_[d] = n; }
        void setNWhispers(int d, int n){ NWhispers_[d] = n; }
        
        // <初期化>
        void initMatch(const MatchRule& rule)noexcept{
            rule_ = rule;
        }
        void clear()noexcept{ // データを追加できる状態まで消去する
            rule_.clear();
            
            game_ = day_ = 0;
            talkIndex_ = whisperIndex_ = whisperTurn_ = 0;
            
            attack_.clear(); divine_.clear(); guard_.clear();
            executed_.clear(); attacked_.clear(); divined_.clear(); inquested_.clear();
            actionFlag.clear();
            
            deathDay_.clear(); causeOfDeath_.clear();
            
            playerRole_.clear();
            rolePlayers_.fill(0);
            
            for(int d = 0; d < N_MAX_DAYS; ++d){
                alive_[d].reset();
                for(int r = 0; r < N_ROLES; ++r){
                    roleAlive_[d][r].reset();
                }
                vote_[d].clear();
                attackVote_[d].clear();
            }
            for(int d = 0; d < N_MAX_DAYS; ++d){
                for(int i = 0; i < N_TALK_TURNS * N_MAX_PLAYERS; ++i){
                    talk_[d][i] = Talk(0);
                }
            }
            for(int d = 0; d < N_MAX_DAYS; ++d){
                NTalks_[d] = 0;
            }
            for(int d = 0; d < N_MAX_DAYS; ++d){
                for(int i = 0; i < N_TALK_TURNS * N_WHISPER_TURNS * maxPlayerNum(Role::WEREWOLF); ++i){
                    whisper_[d][i] = Whisper(0);
                }
            }
            for(int d = 0; d < N_MAX_DAYS; ++d){
                NWhispers_[d] = 0;
                for(int t = 0; t < N_TALK_TURNS + 1; ++t){
                    TurnWhisperFirstIndex_[d][t] = 0;
                }
            }
        }
        void initGame()noexcept{
            day_ = talkIndex_ = 0;
      
            attack_.clear(); divine_.clear(); guard_.clear();
            executed_.clear(); attacked_.clear(); divined_.clear(); inquested_.clear();
            actionFlag.clear();
            
            deathDay_.clear(); causeOfDeath_.clear();
            
            playerRole_.fill(NONE_ROLE);
            rolePlayers_.fill(0);
            
            for(int d = 0; d < N_MAX_DAYS; ++d){
                alive_[d].reset();
                for(int r = 0; r < N_ROLES; ++r){
                    roleAlive_[d][r].reset();
                }
                vote_[d].clear();
                attackVote_[d].clear();
            }
            
            for(int d = 0; d < N_MAX_DAYS; ++d){
                for(int i = 0; i < N_TALK_TURNS * N_MAX_PLAYERS; ++i){
                    talk_[d][i] = Talk(0);
                }
            }
            for(int d = 0; d < N_MAX_DAYS; ++d){
                NTalks_[d] = 0;
            }
            for(int d = 0; d < N_MAX_DAYS; ++d){
                for(int i = 0; i < N_TALK_TURNS * N_WHISPER_TURNS * maxPlayerNum(Role::WEREWOLF); ++i){
                    whisper_[d][i] = Whisper(0);
                }
            }
            for(int d = 0; d < N_MAX_DAYS; ++d){
                NWhispers_[d] = 0;
                for(int t = 0; t < N_TALK_TURNS + 1; ++t){
                    TurnWhisperFirstIndex_[d][t] = 0;
                }
            }

        }
        void initWhisper()noexcept{
            whisperTurn_ = 0;
        }
        void closeWhisper()noexcept{
            ++whisperTurn_;
        }
        
        void update(){ // 更新
        }
        
        void proceedDay(int d){
            // 日付が変わり、情報更新を受けた後の準備(json用, 思考用)
            if(d > DAY_FIRST){
                // execute と attackの情報によりaliveを更新
                for(int r = 0; r < N_ROLES; ++r){
                    roleAlive_[d][r] = roleAlive_[d - 1][r];
                }
                alive_[d] = alive_[d - 1];
                if(anyExecuted(d - 1)){
                    int e = executed(d - 1);
                    alive_[d].reset(e);
                    if(playerRole(e) != NONE_ROLE){
                        roleAlive_[d][playerRole(e)].reset(e);
                    }
                }
                if(anyAttacked(d - 1)){
                    int a = attacked(d - 1);
                    alive_[d].reset(a);
                    if(playerRole(a) != NONE_ROLE){
                        roleAlive_[d][playerRole(a)].reset(a);
                    }
                }
            }else{
                for(int r = 0; r < N_ROLES; ++r){
                    roleAlive_[0][r] = rolePlayers(r);
                }
                alive_[0].fill(0, rule().players - 1);
            }
        }
        
        // <サーバーからの情報更新>
        void setTalkIndex(int d, int i)noexcept{
            talkIndex_ = i;
            NTalks_[d] = i; // その日ここまでの会話数
        }
        
        bool exam()const{
            // 状態
            // aliveな人数
            for(int r = 0; r < N_ROLES; ++r){
                if(rule().rolePlayers[r] < rolePlayers(r).count()){
                    cerr << "Info::exam() : more role players than rule. " << roleChar[r] << endl; return false;
                }
            }
            for(int d = DAY_FIRST; d <= day(); ++d){
                for(int r = 0; r < N_ROLES; ++r){
                    if(rule().rolePlayers[r] < roleAlive(d, r).count()){
                        cerr << "Info::exam() : more alive role players than rule. " << roleChar[r] << endl; return false;
                    }
                    if(!holdsBits(alive(d), roleAlive(d, r))){
                        cerr << "Info::exam() : alives hold error. " << roleChar[r] << endl; return false;
                    }
                    
                }
            }
            
            
            // 行動フラグ
            
            return true;
        }
        
        
        
        std::string toString()const{
            std::ostringstream oss;
            oss << rule().toString();
            oss << "game = " << game() << " day = " << day();
            oss << " talkIndex = " << talkIndex() << " whisperTurn = " << whisperIndex() << endl;
            oss << "alive = " << alive(day()).toSetString() << endl;
            oss << "executed  = " << executedAll() << endl;
            oss << "attack    = " << attackAll() << endl;
            oss << "attacked  = " << attackedAll() << endl;
            oss << "guard     = " << guardAll() << endl;
            oss << "divine    = " << divineAll() << endl;
            oss << "divined   = " << divinedAll() << endl;
            oss << "inquested = " << inquestedAll() << endl;
            oss << "playerRole = " << playerRole_ << endl;
            oss << "vote, attackVote = " << endl;
            for(int d = 0; d < N_MAX_DAYS; ++d){
                oss << voteDay(d) << " " << attackVoteDay(d) << endl;
            }
            oss << "talks = ";
            for(int d = 0; d < N_MAX_DAYS; ++d){
                oss << NTalks(d) << ", ";
            }oss << endl;
            oss << "whispers = ";
            for(int d = 0; d < N_MAX_DAYS; ++d){
                oss << NWhispers(d) << ", ";
            }oss << endl;
            return oss.str();
        }
        
        template<class callback_t>
        void iterateConversation(const callback_t& callback){
            
        }
        
        std::string toConversationString(int ds = -1)const{
            if(ds < 0){ ds = day() + 1; }
            std::ostringstream oss;
            for(int d = 0; d < ds; ++d){
                /*int ti = 0, wi = 0;
                 while(ti < )
                 for(int t = 0; t < rule.turns; ++t, ++ti){
                 oss << "Talk : ";
                 for(int i = 0; i < ANlive(0); ++i){
                 int di = i + NAlive(d) * t;
                 }
                 }
                 oss << ""*/
                oss << "Day : " << d << endl;
                oss << "Talk :";
                for(int i = 0; i < NTalks(d); ++i){
                    oss << " [" << talk(d, i).player() << "]" << talk(d, i);
                }oss << endl;
                oss << "Day : " << d << endl;
                oss << "Whisper :";
                for(int i = 0; i < NWhispers(d); ++i){
                    oss << " [" << whisper(d, i).player() << "]" << whisper(d, i);
                }oss << endl;
            }
            return oss.str();
        }
        
        std::string toAllString(int d = -1)const{
            return toString() + toConversationString(d);
        }
    };
    
    struct AgentInfo : public Info{ // プレーヤー視点の情報
        int myIndex_; // 自分のプレーヤー番号
        
        int myIndex()const noexcept{ return myIndex_; }
        void setMyIndex(int i)noexcept{ myIndex_ = i; }
  
        std::string toString()const{
            std::ostringstream oss;
            oss << "myIndex = " << myIndex() << endl;
            oss << Info::toString();
            return oss.str();
        }
    };
    
    
    
    /**************************試合ログ**************************/
    
    struct Log : public Info{
        std::array<char[32], N_MAX_PLAYERS> name_; // プレーヤーの名前
        int winner_; // 勝利チーム
        int days_; // ゲーム全体の日数
        char fileName_[128]; // ログファイルの名前
        std::string name(int p)const{
            return std::string(name_[p]);
        }
        std::string fileName()const{
            return std::string(fileName_);
        }
        void setName(int p, const std::string& nm){
            snprintf(&name_[p][0], 32, nm.c_str());
        }
        void setFileName(const std::string& fn){
            snprintf(fileName_, 128, fn.c_str());
        }
        int winner()const noexcept{
            return winner_;
        }
        int days()const noexcept{
            return days_;
        }
        void setWinner(int d, int t)noexcept{
            days_ = d;
            winner_ = t;
        }
        void clear(){
            Info::clear();
            winner_ = -1;
            for(int p = 0; p < N_MAX_PLAYERS; ++p){
                snprintf(&name_[p][0], 32, "");
            }
            snprintf(fileName_, 128, "");
        }
        
        std::string toString(int ds = -1)const{
            if(ds < 0){ ds = days(); }
            std::ostringstream oss;
            oss << fileName() << " (" << days() << " days, ";
            oss << toTeamString(winner_) << " won)" << endl;
            return oss.str() + Info::toString();
        }
        std::string toAllString(int ds = -1)const{
            if(ds < 0){ ds = days(); }
            std::ostringstream oss;
            oss << fileName() << " (" << days() << " days, ";
            oss << toTeamString(winner_) << " won)" << endl;
            return oss.str() + Info::toAllString(ds);
        }
        
        bool examFilled()const{
            // 全ての情報を読み込めたか確認
            if(!examTeam(winner())){ return false; }
            return true;
        }
    };
    
    /**************************多次元確率分布の演算**************************/
    
    std::ostream& operator<<(std::ostream& os, const double pd[N_MAX_PLAYERS][N_ROLES]){
        for(int i = 0; i < N_MAX_PLAYERS; ++i){
            for(int j = 0; j < N_ROLES; ++j){
                os << " " << pd[i][j];
            }os << endl;
        }
        return os;
    }
    
    void normalize2(double pd[N_MAX_PLAYERS][N_ROLES], int R, const MatchRule& rule){
        const int I = rule.players;
        const int J = N_ROLES;
        for(int r = 0; r < R; ++r){
            // i方向で正規化
            for(int j = 0; j < J; ++j){
                double sum = 0;
                for(int i = 0; i < I; ++i){
                    sum += pd[i][j];
                }
                if(sum > 0){
                    for(int i = 0; i < I; ++i){
                        pd[i][j] *= rule.rolePlayers[j] / sum;
                    }
                }else{ // たまたま0になってしまった場合、全て等確率に戻す
                    for(int i = 0; i < I; ++i){
                        pd[i][j] = rule.rolePlayers[j] / double(I);
                    }
                }
            }
            // j方向で正規化
            for(int i = 0; i < I; ++i){
                double sum = 0;
                for(int j = 0; j < J; ++j){
                    sum += pd[i][j];
                }
                //cerr << "sum = " << sum << endl;
                if(sum > 0){
                    for(int j = 0; j < J; ++j){
                        pd[i][j] *= 1 / sum;
                    }
                }else{ // たまたま0になってしまった場合、全て等確率に戻す
                    for(int j = 0; j < J; ++j){
                        pd[i][j] = 1 / double(J);
                    }
                }
            }
        }
    }
    void normalize2(double pd[N_MAX_PLAYERS][N_ROLES], int R,
                    const int I, const BitArray64<4, 16> seq, const BitArray64<8, 8> rolePlayerNum){
        // プレーヤーを指定して正規化
        // 各役職の残り人数も指定
        const int J = N_ROLES;
        for(int r = 0; r < R; ++r){
            //cerr << pd;
            
            // i方向で正規化
            for(int j = 0; j < J; ++j){
                double sum = 0;
                for(int i = 0; i < I; ++i){
                    sum += pd[seq[i]][j];
                }
                if(sum > 0){
                    for(int i = 0; i < I; ++i){
                        pd[seq[i]][j] *=  rolePlayerNum[j] / sum;
                    }
                }else{ // たまたま0になってしまった場合、全て等確率に戻す
                    for(int i = 0; i < I; ++i){
                        pd[seq[i]][j] = rolePlayerNum[j] / double(I);
                    }
                }
            }
            // j方向で正規化
            for(int si = 0; si < I; ++si){
                int i = seq[si];
                double sum = 0;
                for(int j = 0; j < J; ++j){
                    sum += pd[i][j];
                }
                if(sum > 0){
                    for(int j = 0; j < J; ++j){
                        pd[i][j] *= 1 / sum;
                    }
                }else{ // たまたま0になってしまった場合、全て等確率に戻す
                    for(int j = 0; j < J; ++j){
                        pd[i][j] = 1 / double(J);
                    }
                }
                //ASSERT(sum > 0, cerr << rolePlayerNum << endl << pd);
            }
        }
    }
    
    template<class dice_t>
    BitArray64<4, 16> dealRolePlain(double pd[N_MAX_PLAYERS][N_ROLES], const MatchRule& rule, dice_t *const pdice){
        // 役職分配
        
        const int P = rule.players;
        BitArray64<4, 16> ret(0);
        
        BitArray64<4, 16> seq(0);
        BitArray64<8, 8> roleNum = rule.rolePlayers;
        
        // まだ不明なプレーヤーのビット配列
        for(int p = 0; p < P; ++p){
            seq.set(p, p);
        }
        // 多次元確率分布に従って役を分配
        // 役の決定順をランダムに並べ替え
        for(int n = P - 1; n > 0; --n){
            int r = pdice->rand() % (n + 1);
            seq.swap(n, r);
        }
        for(int n = 0; n < P - 1; ++n){
            // 正規化
            normalize2(pd, 4, P - n, seq, roleNum);
            
            for(int p = 0; p < P; ++p){
                for(int r = 0; r < N_ROLES; ++r){
                    //cerr << " " << pd[p][r];
                    if(roleNum[r] == 0 && pd[p][r] > 0){
                        //cerr << "zero but non-zero" << pd[p][r] << endl;
                    }
                }//cerr << endl;
            }
            
            // seq[0] の役職を決定
            const int p = seq[0];
            double dr = pdice->drand() * 0.999999999999; // 正規化の丸め誤差にかからないように
            int role = N_ROLES - 1;
            for(int r = 0; r < N_ROLES - 1; ++r){
                dr -= pd[p][r];
                if(dr <= 0){ role = r; break; }
            }
            
            ASSERT(!ret[p], cerr << ret << endl;);
            
            ret.set(p, role);
            seq.remove(0); // seqの1人目を外して詰める
            roleNum.subtr(role, 1); // 決定した役職を一人減らす
            
            // 以下デバッグ用に綺麗にしておく
            for(int r = 0; r < N_ROLES; ++r){
                pd[p][r] = 0;
            }
            //pd[p][role] = 1;
        }
        // 最後の一人を決定
        //cerr << seq << roleNum << endl;
        int lp = seq[0];
        int lr = bsf64(roleNum) / 8;
        //cerr << lp << " " << lr << endl;
        ret.set(lp, lr);
        
        return ret;
    }
    
    template<int R, class info_t, class rp_t, class rn_t>
    BitArray64<4, 16> prepareWithCertainInfo(double pd[N_MAX_PLAYERS][N_ROLES], rp_t& ret,
                                             rn_t& roleNum, const info_t& info, const int myId){
        // 確実な情報によってpdをmodify、かつroleNumから引いておき、uncertain seqを返す
        auto alive = info.alive(DAY_FIRST);
        const auto& rule = info.rule();
        const int P = rule.players;
        
        switch(R){
            case Role::VILLAGER: break; // 村人は自分のみ
            case Role::SEER: { // 占い師は占い結果を利用
                for(int p = 0; p < P; ++p){ pd[p][R] = 0; }
                for(int d = DAY_DIVINE_FIRST; d < info.day(); ++d){
                    if(info.anyDivine(d, myId)){ // 占いを行った
                        int p = info.divine(d);
                        if(info.divined(d) == Species::WEREWOLF && ret[p] != Role::WEREWOLF){ // 人狼判定(まだ出ていない)
                            ret.set(p, Role::WEREWOLF); alive.reset(p);
                            roleNum.subtr(Role::WEREWOLF, 1);
                            if(roleNum[Role::WEREWOLF] == 0){ // 人狼が全員判明した場合はそれ以外のプレーヤーの人狼率を0に
                                for(int p = 0; p < P; ++p){
                                    if(ret[p] != Role::WEREWOLF){
                                        pd[p][Role::WEREWOLF] = 0;
                                    }
                                }
                            }
                        }else{ // 人狼でない
                            // 人狼の確率を0にする
                            pd[p][Role::WEREWOLF] = 0;
                        }
                    }
                }
            }break;
            case Role::MEDIUM: { // 霊媒師は霊媒結果を利用
                for(int p = 0; p < P; ++p){ pd[p][R] = 0; }
                for(int d = DAY_EXECUTE_FIRST; d < info.day(); ++d){
                    int p = info.executed(d); // 必ず処刑は起こるので確認の必要なし
                    if(info.inquested(d) == Species::WEREWOLF){ // 人狼判定
                        ret.set(p, Role::WEREWOLF); alive.reset(p);
                        roleNum.subtr(Role::WEREWOLF, 1);
                        // 人狼が全員判明することはない
                    }else{ // 人狼でない
                        // 人狼の確率を0にする
                        pd[p][Role::WEREWOLF] = 0;
                    }
                }
            }break;
            case Role::BODYGUARD:{ // 狩人は自分のみ
                for(int p = 0; p < P; ++p){ pd[p][R] = 0; }
            }break;
            case Role::POSSESSED:{ // 狂人は自分のみ
                for(int p = 0; p < P; ++p){ pd[p][R] = 0; }
            }break;
            case Role::WEREWOLF:{ // 人狼は味方と、単独襲撃失敗の場合に狩人でないことがわかる
                for(int p = 0; p < P; ++p){ pd[p][R] = 0; }
                auto w = info.roleAlive(DAY_FIRST, Role::WEREWOLF);
                ASSERT(w.count() == info.rule().rolePlayers[Role::WEREWOLF], cerr << w << endl << info.toString(););
                iterate(w, [&ret, &alive, &pd](int p)->void{
                    ret.set(p, Role::WEREWOLF); alive.reset(p);
                    for(int r = 0; r < N_ROLES; ++r){ pd[p][r] = 0; }
                    pd[p][R] = 1;
                });
                roleNum.assign(R, 0);
                
                for(int d = DAY_ATTACK_FIRST; d < info.day(); ++d){
                    if(info.anyAttack(d) && !info.anyAttacked(d)){ // 襲撃にトライしたが失敗
                        // そのときにすでに死んでいたプレーヤーは狩人ではない
                        for(int p = 0; p < P; ++p){
                            if(!info.isAliveAfterExecution(d, p)){
                                pd[p][Role::BODYGUARD] = 0;
                            }
                        }
                    }
                }
                
            }break;
            default: break;
        }
        
        // 役職によってすでにわかっている情報から決定
        if(R != NONE_ROLE){
            if(R != Role::WEREWOLF){
                ret.set(myId, R); alive.reset(myId); roleNum.subtr(R, 1); // 自分
                for(int r = 0; r < N_ROLES; ++r){ pd[myId][r] = 0; }
                pd[myId][R] = 1;
            }
        }
        
        // まだ不明なプレーヤーのビット配列
        return alive.template toArray64<4, 16>();
    }
    
    template<class info_t, class rp_t, class rn_t>
    BitArray64<4, 16> prepareWithCertainInfoWithRole(double pd[N_MAX_PLAYERS][N_ROLES], rp_t& ret,
                                             rn_t& roleNum, const info_t& info, const int myId, const int myRole){
        switch(myRole){
            case 0: return prepareWithCertainInfo<0>(pd, ret, roleNum, info, myId); break;
            case 1: return prepareWithCertainInfo<1>(pd, ret, roleNum, info, myId); break;
            case 2: return prepareWithCertainInfo<2>(pd, ret, roleNum, info, myId); break;
            case 3: return prepareWithCertainInfo<3>(pd, ret, roleNum, info, myId); break;
            case 4: return prepareWithCertainInfo<4>(pd, ret, roleNum, info, myId); break;
            case 5: return prepareWithCertainInfo<5>(pd, ret, roleNum, info, myId); break;
            default: return prepareWithCertainInfo<NONE_ROLE>(pd, ret, roleNum, info, myId); break;
        }
    }
    
    template<int R, class info_t, class dice_t>
    BitArray64<4, 16> dealRole(double pd[N_MAX_PLAYERS][N_ROLES],
                               const info_t& info, const int myId, dice_t *const pdice){
        // プレーヤーの持つ情報を用いた役職分配
        //cerr << "my index = " << myId << endl;
        BitArray64<4, 16> ret(0);
        BitArray64<8, 8> roleNum = info.rule().rolePlayers;
        
        // 確定情報の扱い
        BitArray64<4, 16> seq = prepareWithCertainInfo<R>(pd, ret, roleNum, info, myId);
        const int P = roleNum.sum();
        
        // 変数の値チェック
        for(int r = 0; r < N_ROLES; ++r){
            ASSERT(roleNum[0] <= info.rule().rolePlayers,);
            ASSERT(P == roleNum.sum(),);
        }
        
        // 多次元確率分布に従って役を分配
        
        // 役の決定順をランダムに並べ替え
        for(int n = P - 1; n > 0; --n){
            int r = pdice->rand() % (n + 1);
            seq.swap(n, r);
        }
        //cerr << seq << roleNum << endl;
        
        for(int n = 0; n < P - 1; ++n){
            // 正規化
            normalize2(pd, 4, P - n, seq, roleNum);
            
            /*for(int p = 0; p < P; ++p){
                for(int r = 0; r < N_ROLES; ++r){
                    cerr << " " << pd[p][r];
                    if(roleNum[r] == 0 && pd[p][r] > 0){
                        cerr << "zero but non-zero" << pd[p][r] << endl;
                    }
                }cerr << endl;
            }
            getchar();*/
            
            // seq[0] の役職を決定
            const int p = seq[0];
            double dr = pdice->drand() * 0.999999999999; // 正規化の丸め誤差にかからないように
            int role = N_ROLES - 1;
            for(int r = 0; r < N_ROLES - 1; ++r){
                dr -= pd[p][r];
                if(dr <= 0){ role = r; break; }
            }
            ret.set(p, role);
            seq.remove(0); // seqの1人目を外して詰める
            roleNum.subtr(role, 1); // 決定した役職を一人減らす
            
            // 以下デバッグ用に綺麗にしておく
            for(int r = 0; r < N_ROLES; ++r){
                pd[p][r] = 0;
            }
            //pd[p][role] = 1;
        }
        // 最後の一人を決定
        //cerr << seq << roleNum << endl;
        int lp = seq[0];
        int lr = bsf64(roleNum) / 8;
        //cerr << lp << " " << lr << endl;
        ret.set(lp, lr);
        
        return ret;
    }
    
    template<class info_t, class dice_t>
    BitArray64<4, 16> dealRoleSubjective(double pd[N_MAX_PLAYERS][N_ROLES],
                                         const info_t& info, const int subject, dice_t *const pdice){
        int r = NONE_ROLE;
        if(examAgentNum(subject)){
            r = info.playerRole(subject);
        }
        BitArray64<4, 16> pr;
        switch(r){
            case 0: pr = dealRole<0>(pd, info, subject, pdice); break;
            case 1: pr = dealRole<1>(pd, info, subject, pdice); break;
            case 2: pr = dealRole<2>(pd, info, subject, pdice); break;
            case 3: pr = dealRole<3>(pd, info, subject, pdice); break;
            case 4: pr = dealRole<4>(pd, info, subject, pdice); break;
            case 5: pr = dealRole<5>(pd, info, subject, pdice); break;
            default: pr = dealRole<6>(pd, info, -1, pdice); break;
        }
        return pr;
    }
    
    int countNCorrectRoles(BitArray64<4, N_MAX_PLAYERS> a,
                     BitArray64<4, N_MAX_PLAYERS> b,
                     int P){
        // 同じ役職の人数を数える
        int cnt = 0;
        for(int p = 0; p < P; ++p){
            if(a[p] == b[p]){ ++cnt; }
        }
        return cnt;
    }
    
    template<class info_t>
    std::pair<int, int> certainRoleOrSpecies(const info_t& info, int myId, int player, int R){
        // 主観的に役職もしくは種族が確定する場合
        // 速度超遅なので末端で呼ぶのは禁止
        std::pair<int, int> ret = std::make_pair(NONE_ROLE, NONE_SPECIES);
        
        if(player == myId){ ret.first = info.playerRole(player); return ret; }

        switch(R){
            case Role::VILLAGER: break; // 村人は自分のみ
            case Role::SEER: { // 占い師は占い結果を利用
                for(int d = DAY_DIVINE_FIRST; d < info.day(); ++d){
                    if(info.anyDivine(d, myId)){ // 占いを行った
                        int p = info.divine(d);
                        if(p == player){
                            if(info.divined(d) == Species::WEREWOLF){ // 人狼判定
                                ret.first = Role::WEREWOLF; return ret;
                            }else{ // 人狼でない
                                ret.second = Species::HUMAN; return ret;
                            }
                        }
                    }
                }
            }break;
            case Role::MEDIUM: { // 霊媒師は霊媒結果を利用
                if(info.causeOfDeath(player) == EXECUTED){
                    for(int d = DAY_EXECUTE_FIRST; d < info.day(); ++d){
                        int p = info.executed(d); // 必ず処刑は起こるので確認の必要なし
                        if(p == player){
                            if(info.inquested(d) == Species::WEREWOLF){ // 人狼判定
                                ret.first = Role::WEREWOLF; return ret;
                            }else{ // 人狼でない
                                ret.second = Species::HUMAN; return ret;
                            }
                        }
                    }
                }
            }break;
            case Role::BODYGUARD: break; // 狩人は自分のみ
            case Role::POSSESSED: break; // 狂人は自分のみ
            case Role::WEREWOLF:{ // 人狼は味方がわかる
                auto w = info.roleAlive(DAY_FIRST, Role::WEREWOLF);
                if(w.test(player)){
                    ret.first = Role::WEREWOLF; return ret;
                }
            }break;
            default: break;
        }
        return ret;
    }
    
    template<int D, int R, class move_t, class info_t>
    int genValidActionsTemp(move_t *const mv0, const info_t& info, int player){
        // まともな行動を全て生成
        // 日によって行動内容が違うので注意
        // TODO : 行動がない日は呼ばれないようにしないと遅い
        int day = info.day();
        move_t *mv = mv0;
        
        if(D >= DAY_EXECUTE_FIRST){
            // 生存している自分以外に投票できる
            //cerr << endl;
            iterateExcept(info.alive(day), player, [&mv, &info, player, day](int p0){
                
                // 村人陣営で白の場合と、人狼陣営で黒の場合には候補から外す
                if(info.sub[player].certainPlayerSpecies[p0] != NONE_SPECIES){
                    if(getTeam(R) == Team::VILLAGER
                       && info.sub[player].certainPlayerSpecies[p0] == Species::HUMAN){
                        return;
                    }else if(getTeam(R) == Team::WEREWOLF
                             && info.sub[player].certainPlayerSpecies[p0] == Species::WEREWOLF){
                        return;
                    }
                }
                
                // 占い師で黒判定を出せている場合には、それ以外のプレーヤーは投票対象にしない
                if(R == Role::SEER){
                    bool black = false;
                    for(int p = 0; p < info.rule().players; ++p){
                        if(info.isAlive(day, p) && p != player){
                            if(info.sub[player].certainPlayerSpecies[p] == Species::WEREWOLF){
                                black = true; break;
                            }
                        }
                    }
                    if(black){
                        if(info.sub[player].certainPlayerSpecies[p0] != Species::WEREWOLF){
                            return;
                        }
                    }
                }
                
                //cerr << p0 << " ";
                switch(R){
                    case Role::SEER:{ // 種族未確定の相手を占う
                        int cnt = 0;
                        if(D >= DAY_DIVINE_FIRST){
                            //cerr << info.sub[player].certainPlayerSpecies << endl;
                            for(int p1 = 0; p1 < info.rule().players; ++p1){
                                if(info.isAlive(day, p1) && p1 != player){
                                    if(info.sub[player].certainPlayerSpecies[p1] == NONE_SPECIES){ // 種族未確定
                                        mv->clear(); mv->set(0, p0); mv->set(1, p1); ++mv; ++cnt;
                                    }
                                }
                            }
                        }
                        if(cnt == 0){ // 占いは適当な相手に限定
                            //cerr << "no kouho" << endl;
                            //auto alive = info.alive(day);
                            //alive.reset(player);
                            //mv->clear(); mv->set(0, p0); mv->set(1, alive.bsf()); ++mv;
                            mv->clear(); mv->set(0, p0); mv->set(1, player); ++mv;
                        }
                    }break;
                    case Role::BODYGUARD: { // 自分以外を護衛できる
                        if(D >= DAY_ATTACK_FIRST){
                            iterateExcept(info.alive(day), player, [&mv, &info, p0](int p1){
                                mv->clear(); mv->set(0, p0); mv->set(1, p1); ++mv;
                            });
                        }
                    }break;
                    case Role::WEREWOLF: { // 人狼以外を襲撃でき、棄権もできる（が未実装）
                        if(D >= DAY_ATTACK_FIRST){
                            auto can = info.speciesAlive(day, Species::HUMAN);
                            iterateExcept(can, player, [&mv, p0](int p1){
                                mv->clear(); mv->set(0, p0); mv->set(1, p1); ++mv;
                            });
                        }
                    }break;
                    default: { // 処刑投票だけのプレーヤー
                        mv->clear(); mv->set(0, p0);  ++mv;
                    }break;
                }
            });
        }else{
            if(D >= DAY_DIVINE_FIRST && R == Role::SEER){ // 種族未確定の相手を占う
                for(int p1 = 0; p1 < info.rule().players; ++p1){
                    if(p1 != player){
                    //if(info.sub[player].certainPlayerSpecies[p1] == NONE_SPECIES){ // 種族未確定
                        //if(p1 != player){ mv->clear(); mv->set(1, p1); ++mv; }
                        mv->clear(); mv->set(1, p1); ++mv;
                    //}
                    }
                }
            }
        }
        
        // 生成着手のチェック
        for(const move_t *tmv = mv0; tmv != mv; ++tmv){
            ASSERT(info.isValidAction(day, player, *tmv),
                   cerr << player << " " << roleChar[R] << " " << (*tmv) << endl;
                   cerr << info.toString() << endl;);
        }
        
        return mv - mv0; // 行動候補の総数を返す
    }
    
    template<class move_t, class info_t>
    int genValidActions(move_t *const mv0, const info_t& info, int player){
        const unsigned int day = info.day();
        switch(info.playerRole(player)){
            case 0: if(day > 0){ return genValidActionsTemp<1, 0>(mv0, info, player);
            }else{ return genValidActionsTemp<0, 0>(mv0, info, player); } break;
            case 1: if(day > 0){ return genValidActionsTemp<1, 1>(mv0, info, player);
            }else{ return genValidActionsTemp<0, 1>(mv0, info, player); } break;
            case 2: if(day > 0){ return genValidActionsTemp<1, 2>(mv0, info, player);
            }else{ return genValidActionsTemp<0, 2>(mv0, info, player); } break;
            case 3: if(day > 0){ return genValidActionsTemp<1, 3>(mv0, info, player);
            }else{ return genValidActionsTemp<0, 3>(mv0, info, player); } break;
            case 4: if(day > 0){ return genValidActionsTemp<1, 4>(mv0, info, player);
            }else{ return genValidActionsTemp<0, 4>(mv0, info, player); } break;
            case 5: if(day > 0){ return genValidActionsTemp<1, 5>(mv0, info, player);
            }else{ return genValidActionsTemp<0, 5>(mv0, info, player); } break;
            default: UNREACHABLE; break;
        }
        UNREACHABLE;
    }
    
    //double joint
    
    /**************************基本ルール**************************/
    
    constexpr MatchRule BASE_RULE = MatchRule(15, 10,
                                             (8ULL << (8 * Role::VILLAGER))
                                             | (1ULL << (8 * Role::SEER))
                                             | (1ULL << (8 * Role::MEDIUM))
                                             | (1ULL << (8 * Role::BODYGUARD))
                                             | (1ULL << (8 * Role::POSSESSED))
                                             | (3ULL << (8 * Role::WEREWOLF)));
}

#endif // AIWOLF_AIWOLF_HPP_