/*
 eval_learner.cc
 Katsuki Ohto
 */

// 評価関数の学習
// 1日の内容が全て終わった時点を想定

#include "../aiwolf.hpp"
#include "../thousand/evaluation.hpp"

using namespace AIWolf;
using namespace AIWolf::Thousand;

enum{
    MODE_FLAG_TEST = 1,
    MODE_FLAG_SHUFFLE = 2,
};

int N = 10000;

template<class log_t, class evaluator_t, class learner_t>
int learnEvalParamsDayStart(const log_t& log, const BitSet32 flags,
                            const evaluator_t& eval, learner_t *const plearner){

    double score[N_TEAMS] = {0};
    ThinkInfo info(log);
    info.clearConversation();
    
    for(int d = 0; d < log.days(); ++d){
        if(log.isGameOver(d) != NONE_TEAM){ break; }
        
        info.proceedConversation(d); // 情報を更新
        info.addConversationActionRelation(d); // 宣言-行動関係を更新

        info.setDay(d + 1);
        
        // d + 1日目の朝の状態を評価する
        int idx = log.winner();
        
        calcEvaluationScore<1>(&score[0], info, eval);
        
        if(flags.test(1)){ // feed feature value
            plearner->feedFeatureValue();
        }
        if(flags.test(0)){ // learn
            plearner->acceptDecidedMoveIndex(idx);
            plearner->learnByMove();
        }
        if(flags.test(2)){ // test
            plearner->feedObjValue(idx);
        }
    }
    return 0;
}

Evaluator eval;
EvaluatorLearner learner;

int learn(const std::string& lpath, const std::string& opath, int mode, int iterations){
    
    double learnRate = (mode & MODE_FLAG_TEST) ? 0.8 : 1;
    
    std::mt19937 mt((uint32)time(NULL));
    
    eval.setLearner(&learner);
    learner.setPolicy(&eval);
    
    // ログをファイルから読み込み
    Log *const log = new Log[N];
    FILE *const pf = fopen(lpath.c_str(), "rb");
    fread(log, sizeof(Log) * N, 1, pf);
    fclose(pf);
    
    cerr << N << " games were imported." << endl;
    
    // preparing
    learner.setLearnParam(1, 0.00005, 0, 0.0000002, 64);

    const int games = N;
    const int learnGames = min((int)(games * learnRate), games);
    const int testGames = games - learnGames;
    
    cerr << learnGames << " games for learning, " << testGames << " games for test." << endl;
    
    // 特徴要素の解析(学習時のステップ幅決め)
    cerr << "started analyzing feature." << endl;
    
    BitSet32 flag(0);
    flag.set(1);
    for(int i = 0; i < games; ++i){
        learnEvalParamsDayStart(log[i], flag, eval, &learner); // 特徴要素解析
    }
    learner.closeFeatureValue();
    learner.foutFeatureSurvey(opath + "eval_feature_survey.dat");
    learner.finFeatureSurvey(opath + "eval_feature_survey.dat");
    cerr << "record : " << learner.toRecordString() << endl;
    
    // learning
    cerr << "started learning." << endl;
    
    std::vector<int> num;
    num.reserve(N);
    for(int i = 0; i < N; ++i){ num.push_back(i); }
    
    for (int j = 0; j < iterations; ++j){
        
        cerr << "iteration " << j << endl;
        
        learner.setLearnParam(1, 0.00001, 0, 0.0000001, 64);
        
        BitSet32 flag(0);
        flag.set(0);
        flag.set(2);
        
        // 学習フェーズ
        learner.initObjValue();
        std::shuffle(num.begin(), num.begin() + learnGames, mt); // 学習順入れ替え
        for(int i = 0; i < learnGames; ++i){
            learnEvalParamsDayStart(log[num[i]], flag, eval, &learner); // 学習
        }
        eval.hout("../c/param/eval_param.h", "eval_param");
        eval.fout(opath + "eval_param.dat");
        foutComment(eval, opath + "eval_comment.txt");
        cerr << "Learning : " << learner.toObjValueString() << endl;
        
        if(testGames > 0){
            // テストフェーズ
            flag.reset(0);
            learner.initObjValue();
            for(int i = learnGames; i < games; ++i){
                learnEvalParamsDayStart(log[i], flag, eval, &learner); // テスト
            }
            cerr << "Test     : " << learner.toObjValueString() << endl;
        }
    }
    cerr << "finished learning." << endl;
    
    return 0;
}

int main(int argc, char* argv[]){
    
    // 基本的な初期化
    setvbuf(stdout, NULL, _IONBF, 0);
    setvbuf(stdin, NULL, _IONBF, 0);
    setvbuf(stderr, NULL, _IONBF, 0);
    
    std::string lpath = "/Users/ohto/documents/data/aiwolf/dump/binlog.bin";
    std::string opath = "/Users/ohto/documents/data/aiwolf/dump/";
    
    int mode = MODE_FLAG_SHUFFLE;
    int itr = 50;
    
    for(int c = 1; c < argc; ++c){
        if(strstr(argv[c], "-t")){
            mode |= MODE_FLAG_TEST;
        }else if(strstr(argv[c], "-i")){
            itr = atoi(argv[c + 1]);
        }else if(strstr(argv[c], "-l")){
            lpath = std::string(argv[c + 1]);
        }else if(strstr(argv[c], "-n")){ // ログファイルの使用数
            N = atoi(argv[c + 1]);
        }else if(strstr(argv[c], "-o")){
            opath = std::string(argv[c + 1]);
        }else if(strstr(argv[c], "-f")){
            mode &= (~MODE_FLAG_SHUFFLE);
        }
    }

    int ret = learn(lpath, opath, mode, itr);
    
    return ret;
}
