/*
 policy_learner.cc
 Katsuki Ohto
 */

#include "../aiwolf.hpp"

#include "../thousand/singleEstimation.hpp"
#include "../thousand/actionPolicy.hpp"

int N = 10000;

using namespace AIWolf;
using namespace AIWolf::Thousand;

XorShift64 dice;

std::array<std::array<double, 2>, N_ROLES> singleActionSum;
std::array<int, N_ROLES> actionNum;

void initAyalysis(){
    for(int r = 0; r < N_ROLES; ++r){
        singleActionSum[r].fill(0);
    }
    actionNum.fill(0);
}

template<class log_t, class policy_t, class learner_t, class estimator_t>
int learnActionPolicyParams(const log_t& log, const BitSet32 flags, const policy_t& pol, learner_t *const plearner,
                            const estimator_t& est){
    
    const int P = log.rule().players;
    BitArray32<16, 2> mv[N_MAX_PLAYERS * N_MAX_PLAYERS]; // 行動生成バッファ
    double score[N_MAX_PLAYERS * N_MAX_PLAYERS]; // 行動方策得点
    
    ThinkInfo info(log);
    info.clearConversation();
    
    for(int d = 0; d < log.days(); ++d){
        info.proceedConversation(d); // 情報を更新
        info.setDay(d);
        
        // この時点で生きている全員の視点で検証
        BitSet32 alive = info.alive(d).data();
        iterate(alive, [d, &score, &mv, P, flags, &info, &log, &est, &pol, plearner](int subject)->void{
            
            if(flags.test(2)){
                //cerr << log.name(subject) << endl;
            }
            if(log.name(subject).find("StrategyPlayer") == std::string::npos){
                return; // Udon以外飛ばす
            }
            
            const int R = info.playerRole(subject);
            BitArray32<16, 2> logMove(0);
            
            // 役職ごとのログの行動をlogMoveに記録
            logMove.assign(0, log.vote(d, subject)); // 処刑投票は全員
            if(R == Role::SEER){
                logMove.assign(1, log.anyDivine(d, subject) ? log.divine(d) : subject);
            }else if(R == Role::BODYGUARD){
                logMove.assign(1, log.anyGuard(d, subject) ? log.guard(d) : subject);
            }else if(R == Role::WEREWOLF){
                logMove.assign(1, log.attackVote(d, subject));
            }
            
            info.sub[subject].setRole(subject, R, info); // 主観的な情報を設定
            
            // 合法行動を列挙
            const int NMoves = genValidActions(mv, info, subject);
            if(NMoves == 0){ return; } // 行動がない日もある
            
            if(!flags.test(1) && dice.drand() > 0.5){ return; }
            
            // 主観に基づいた全員の役職推定を行う
            estimateNow(&info.sub[subject], subject, info, est);
            
            // 行動得点を計算
            calcActionPolicyScoreWithRole<1>(score, mv, NMoves, info, subject, R, pol);
            
            // ログの行動が含まれているか探す
            int index = -1;
            for(int m = 0; m < NMoves; ++m){
                //cerr << mv[m] << endl;
                if(mv[m].data() == logMove.data()){
                    index = m; break;
                }
            }
            
            if(index == -1){ // 見つからなかった
                
                //cerr << subject << roleChar[R] << " " << logMove << endl << info.toString(); getchar();
                
                if(flags.test(1)){ // feed feature value
                    plearner->feedUnfoundFeatureValue(R);
                }
                if(flags.test(2)){ // test
                    plearner->feedObjValue(index, R);
                }
                return;
            }else{ // 一致した
                if(NMoves > 0){

                    if(flags.test(1)){ // feed feature value
                        plearner->feedFeatureValue(R);
                    }
                    if(flags.test(0)){ // learn
                        plearner->acceptDecidedMoveIndex(index, R);
                        plearner->learnByMove();
                    }
                    if(flags.test(2)){ // test
                        plearner->feedObjValue(index, R);
                    }
                }
            }
            
            if(flags.test(2) && !flags.test(0)){ // test onlyのとき
                // 複数の同時行動のうちそれぞれの一致率も計算
                for(int i = 0; i < 2; ++i){
                    double scoreBySingleAction[N_MAX_PLAYERS] = {0};
                    double scoreBySingleActionSum = 0;
                    for(int m = 0; m < NMoves; ++m){
                        scoreBySingleAction[mv[m][i]] += score[m];
                        scoreBySingleActionSum += score[m];
                    }
                    ASSERT(scoreBySingleAction > 0,);
                    // logMoveと一致した分を加算
                    singleActionSum[R][i] += scoreBySingleAction[logMove[i]] / scoreBySingleActionSum;
                }
                actionNum[R] += 1;
            }
        });
        info.addConversationActionRelation(d); // 宣言-行動関係を更新
    }
    return 0;
}

ActionPolicy pol;
ActionPolicyLearner learner;
Estimator est;

int learn(const std::string& lpath, const std::string& opath, const std::string& epath, int mode, int iterations){
    
    double learnRate = (mode == 0) ? 1 : 0.8;
    
    std::mt19937 mt((uint32)time(NULL));
    
    pol.setLearner(&learner);
    learner.setPolicy(&pol);
    
    // ログをファイルから読み込み
    Log *const log = new Log[N];
    FILE *const pf = fopen(lpath.c_str(), "rb");
    fread(log, sizeof(Log) * N, 1, pf);
    fclose(pf);
    
    cerr << N << " games were imported." << endl;
    
    // 学習とテストの試合数決定
    const int games = N;
    const int learnGames = min((int)(games * learnRate), games);
    const int testGames = games - learnGames;
    
    // 推定器のパラメータを読み込み
    est.fin(epath);
    
    // 特徴要素を解析して、学習の度合いを決める
    learner.setLearnParam(1, 0.00002, 0, 0.000002, 16);
    
    cerr << "started analyzing feature." << endl;
    
    learner.initFeatureValue();
    
    BitSet32 flag(0);
    flag.set(1);
    for(int i = 0; i < games; ++i){
        learnActionPolicyParams(log[i], flag, pol, &learner, est); // 特徴要素解析
    }
    learner.closeFeatureValue();
    learner.foutFeatureSurvey(opath + "pol_feature_survey.dat");
    learner.finFeatureSurvey(opath + "pol_feature_survey.dat");
    
    for(int r = 0; r < N_ROLES; ++r){
        cerr << "record by " << roleChar[r] << " : " << learner.toRecordString(r) << endl;
    }
    
    std::vector<int> num;
    num.reserve(N);
    for(int i = 0; i < N; ++i){ num.push_back(i); }
    
    // 学習開始
    cerr << "started learning." << endl;
    
    for (int j = 0; j < iterations; ++j){
        
        cerr << "iteration " << j << endl;
        
        learner.setLearnParam(1, 0.00001 * (0.01 + 0.99 / sqrt(double(j + 1))), 0, 0.0000001 * (0.01 + 0.99 / sqrt(double(j + 1))), 256);
        
        BitSet32 flag(0);
        flag.set(0);
        flag.set(2);
        
        learner.initObjValue();
        std::shuffle(num.begin(), num.begin() + learnGames, mt); // 学習順入れ替え
        for(int i = 0; i < learnGames; ++i){
            learnActionPolicyParams(log[num[i]], flag, pol, &learner, est); // 学習
        }
        pol.hout("../c/param/pol_param.h", "pol_param");
        pol.fout(opath + "pol_param.dat");
        foutComment(pol, opath + "pol_comment.txt");
        for(int r = 0; r < N_ROLES; ++r){
            cerr << "Learning : " << learner.toObjValueString(r) << endl;
        }
        
        if(testGames > 0){
            flag.reset(0);
            learner.initObjValue();
            
            initAyalysis();
            
            for(int i = learnGames; i < games; ++i){
                learnActionPolicyParams(log[i], flag, pol, &learner, est); // テスト
            }
            for(int r = 0; r < N_ROLES; ++r){ // 全ての視点での結果を表示
                cerr << "Test " << roleChar[r] << " : " << learner.toObjValueString(r) << endl;
            }
            cerr << "single action accuracy = " << endl;
            for(int i = 0; i < 2; ++i){
                for(int r = 0; r < N_ROLES; ++r){
                    cerr << std::setw(5) << int(singleActionSum[r][i] / actionNum[r] * 1000);
                }cerr << endl;
            }
        }
    }
    cerr << "finished learning." << endl;
    
    return 0;
}

int main(int argc, char* argv[]){
    
    //基本的な初期化
    setvbuf(stdout, NULL, _IONBF, 0);
    setvbuf(stdin, NULL, _IONBF, 0);
    setvbuf(stderr, NULL, _IONBF, 0);
    
    std::string lpath = "/Users/ohto/documents/data/aiwolf/dump/binlog.bin";
    std::string opath = "/Users/ohto/documents/data/aiwolf/dump/";
    
    std::string epath = "/Users/ohto/documents/data/aiwolf/dump/est_param.dat";
    
    int mode = 0;
    int itr = 50;
    
    dice.srand((unsigned int)time(NULL) ^ 12345);
    
    for(int c = 1; c < argc; ++c){
        if(strstr(argv[c], "-e")){
            epath = std::string(argv[c + 1]);
        }else if(strstr(argv[c], "-l")){
            lpath = std::string(argv[c + 1]);
        }else if(strstr(argv[c], "-n")){
            N = atoi(argv[c + 1]);
        }else if(strstr(argv[c], "-o")){
            opath = std::string(argv[c + 1]);
        }else if(strstr(argv[c], "-t")){
            mode = 1;
        }
    }
    return learn(lpath, opath, epath, mode, itr);
}