/*
 log_reader.cc
 Katsuki Ohto
 */

#include <dirent.h>
#include <fstream>
#include <climits>

#include "../aiwolf.hpp"

using namespace std;

namespace AIWolf{
    
#define ToI(s) (atoi((s).c_str()))
    
    int readAIWolfLog(const string& path, Log *const pinfo){
        // 公式の.log形式のファイルを読み取り
        string str;
        ifstream ifs(path);
        if(!ifs){ return -1; }
        
        MatchRule rule;
        rule.clear();
        rule.turns = N_TALK_TURNS; // 会話系ターン数はログからすぐにわからないので固定
        pinfo->clear();
        
        while(ifs){
            getline(ifs, str, '\n');
            //cerr << str << endl;
            vector<string> v = split(str, ',');
            int d = ToI(v[0]);
            char c = v[1][0];
            if(c == 's'){ // プレーヤー状態
                int p = toMyAgentNumber(ToI(v[2]));
                if(d == 0){ // 初日に役職等不変の情報を設定
                    int r = toRoleNum(v[3]);
                    rule.rolePlayers.add(r, 1); // 役職人数カウント
                    rule.players += 1; // 合計人数カウント
                    pinfo->setRole(p, r); // 役職設定
                    pinfo->initMatch(rule); // ルールを再設定
                    pinfo->setName(p, v[5]); // プレーヤー名
                }
                if(v[4][0] == 'A'){ // ALIVE 生存
                    pinfo->setAlive(d, p);
                }
            }else if(c == 't'){ // talk 表の会話
                int i = ToI(v[2]);
                pinfo->setPlayerTalk(d, i,
                                     toMyAgentNumber(ToI(v[3])),
                                     toStatementData(v[4], TALK));
                pinfo->setNTalks(d, max(i + 1, pinfo->NTalks(d)));
            }else if(c == 'w'){ // whisper 裏の会話
                int i = ToI(v[2]);
                pinfo->setPlayerWhisper(d, i,
                                     toMyAgentNumber(ToI(v[3])),
                                     toStatementData(v[4], WHISPER));
                pinfo->setNWhispers(d, max(i + 1, pinfo->NWhispers(d)));
            }else if(c == 'v'){ // vote 処刑投票
                pinfo->setVote(d,
                               toMyAgentNumber(ToI(v[2])),
                               toMyAgentNumber(ToI(v[3])));
            }else if(c == 'a'){
                if(v[1].size() > 6){ // attackVote 襲撃投票
                    pinfo->setAttackVote(d,
                                         toMyAgentNumber(ToI(v[2])),
                                         toMyAgentNumber(ToI(v[3])));
                }else{
                    int p = toMyAgentNumber(ToI(v[2]));
                    if(v[3][0] == 't'){ // true 襲撃成功
                        pinfo->setAttack(d, p);
                        pinfo->setAttacked(d, p);
                    }else{ // 襲撃が起こらなかった
                        // 先に処刑または護衛成功したのか、襲撃を放棄したのか判断する必要あり
                        if(examAgentNum(p)
                           && pinfo->isAlive(d, p)
                           && pinfo->playerRole(p) == Role::WEREWOLF){ // 護衛
                            pinfo->setAttack(d, p);
                        }else if(pinfo->judgeAttackAbstaination(d)){ // 襲撃放棄
                            pinfo->setAbandonAttack(d);
                        }else{ // 同じ日に処刑された
                            pinfo->setAttack(d, p);
                        }
                        pinfo->setNoAttacked(d); // 襲撃なし
                    }
                }
            }else if(c == 'd'){ // divine 占い
                int p = toMyAgentNumber(ToI(v[2]));
                int t = toMyAgentNumber(ToI(v[3]));
                if(pinfo->isAlive(d, t)){ // 合法な占い(自分も合法)
                    pinfo->setDivine(d, p, t);
                    int s = toSpeciesNum(v[4]);
                    if(examSpecies(s)){ // 有効な占い結果
                        pinfo->setDivined(d, p, t, s);
                    }else{
                        pinfo->setDivined(d, p, t, NONE_SPECIES);
                    }
                }else{
                    pinfo->setAbandonDivine(d, p);
                }
            }else if(c == 'g'){ // guard 護衛
                pinfo->setGuard(d, toMyAgentNumber(ToI(v[2])), toMyAgentNumber(ToI(v[3])));
            }else if(c == 'e'){ // execete 処刑
                int t = toMyAgentNumber(ToI(v[2]));
                pinfo->setExecuted(d, t);
                pinfo->setInquested(d, 0, t, getSpecies(pinfo->playerRole(t))); // プレーヤー番号は関係ないので適当
            }else if(c == 'r'){ // 試合終了。最後の日付が全体日数
                //cerr << str << endl;
                pinfo->setWinner(d, toTeamNum(v[4])); break;
            }
        }
        // 試合終了まで全ての情報を読み込めたかチェック
        if(!pinfo->examFilled()){
            return -1;
        }
        // ログファイル名
        pinfo->setFileName(path);
        
        //cerr << pinfo->talk(0, 0) << endl;
        //cerr << pinfo->whisper(0, 0) << endl;
        
        return 0;
    }
    
#undef ToI
    
    std::vector<std::string> getLogFilePathVector(const std::string& ipath){
        // ログファイルのパス一覧を作る
        std::vector<std::string> logs;
        
        DIR *pdir;
        dirent *pentry;
        
        pdir = opendir(ipath.c_str());
        if (pdir == nullptr){
            return logs;
        }
        do {
            pentry = readdir(pdir);
            if(pentry != nullptr){
                logs.emplace_back(ipath + std::string(pentry->d_name));
            }
        }while(pentry != nullptr);
        return logs;
    }
    
    int makeRoleWinnerLog(const string& ipath, const string& opath){
        // 役職の残り方パターンと勝利チームの関係のログ
        auto logs = getLogFilePathVector(ipath);
        
        if(logs.size() == 0){ return 0; }
        
        ofstream ofs;
        ofs.open(opath, ios::out);
        ofs.close();
        
        ofs.open(opath, ios::app);
        int cnt = 0;
        for(auto& lp : logs){
            // ログファイルを読む
            if(contains(lp, ".log")){
                Log log;
                cerr << lp << endl;
                if(readAIWolfLog(lp, &log) < 0){
                    cerr << "failed to import" << endl;
                    continue;
                } // 読み込み失敗
                for(int d = 0; d < log.days(); ++d){
                    ofs << log.winner();
                    iterateValidRole([&ofs, &log, d](int r){
                        int n = log.NRoleAlive(d, r);
                        ofs << ' ' << n;
                    });
                    ofs << endl;
                }
                ++cnt;
            }
        }
        ofs.close();
        cerr << "imported " << cnt << " logs and exported to role-winner log." << endl;
        return cnt;
    }
    
    int makeBinLog(const string& ipath, const string& opath, int N){
        // 構造体をそのままバイナリ形式で保存するログ
        if(N < 0){ N = INT_MAX; }
        
        vector<string> lps = getLogFilePathVector(ipath);
        
        std::random_device seed;
        std::mt19937 mt(seed() ^ 12345);
        
        std::shuffle(lps.begin(), lps.end(), mt);
        
        Log *log = new Log[lps.size()];
        
        int cnt = 0;
        for(int i = 0; (cnt < N) && (i < lps.size()); ++i){
            if(contains(lps[i], ".log")){
                cerr << lps[i] << endl;
                if(readAIWolfLog(lps[i], &log[cnt]) < 0){
                    cerr << "failed to import" << endl;
                    continue;
                }else{
                    //cerr << log[cnt].toAllString(); getchar();
                    ++cnt;
                }
            }
        }
        N = min(cnt, N);
        
        FILE *pf = fopen(opath.c_str(), "wb");
        fwrite(log, sizeof(Log) * N, 1, pf);
        fclose(pf);
        
        cerr << "saved " << N << " logs." << endl;
        
        delete[] log;
    }
}

int main(int argc, char *argv[]){
    
    string ipath = "/Users/ohto/documents/data/aiwolf/log/";
    
    //string opath = "/Users/ohto/documents/data/aiwolf/dump/rolewinnerlog.dat";
    
    string opath = "/Users/ohto/documents/data/aiwolf/dump/binlog.bin";
    
    int N = 10000;
    
    for (int i = 1; i < argc; ++i){
        if(strstr(argv[i], "-i")){ // aiwolf-log directory path
            ipath = string(argv[i + 1]);
        }else if(strstr(argv[i], "-n")){ // max numbers of log
            N = atoi(argv[i + 1]);
        }else if(strstr(argv[i], "-o")){ // destination path
            opath = string(argv[i + 1]);
        }
    }
    
    
    
    //AIWolf::makeRoleWinnerLog(ipath, opath);
    AIWolf::makeBinLog(ipath, opath, N);
    
    
    return 0;
}
