/*
 thinkInfo.hpp
 Katsuki Ohto
 */

#ifndef AIWOLF_THOUSAND_THINKINFO_HPP_
#define AIWOLF_THOUSAND_THINKINFO_HPP_

// 思考用の構造体
#include "../aiwolf.hpp"


namespace AIWolf{
    namespace Thousand{
        
        /**************************思考用の局面表現**************************/
        
        struct SubjectiveInfo{ // 主観的な情報
            // 役職確定情報
            BitArray64<8, 8> uncertainRoleNum; // 不確定な役職の人数
            BitArray64<4, 16> certainPlayerRole; // 確定した役職
            
            // 主観的種族確定情報
            BitArray32<2, 16> certainPlayerSpecies;
            
            // 主観的役職予測確率テーブル
            double estPD[N_MAX_PLAYERS][N_ROLES];
            
            void clear()noexcept{
                uncertainRoleNum.clear(); certainPlayerRole.clear(); certainPlayerSpecies.clear();
                memset(estPD, 0, sizeof(estPD));
            }
            
            template<class info_t>
            void setRole(int player, const int r, const info_t& info)noexcept{
                clear();
                // uncertain list を更新
                certainPlayerSpecies.fill(NONE_SPECIES);
                certainPlayerSpecies.set(player, getSpecies(r));
                
                if(r == Role::SEER){ // 占い師で、占いを行っている場合には、その結果を利用
                    for(int d = DAY_DIVINE_FIRST; d < info.day(); ++d){
                        if(info.anyDivine(d, player)){
                            certainPlayerSpecies.assign(info.divine(d), info.divined(d));
                        }
                    }
                    //cerr << certainPlayerSpecies << endl; getchar();
                    
                }else if(r == Role::MEDIUM){ // 霊媒師の場合は必ず霊媒結果を取得する
                    for(int d = DAY_EXECUTE_FIRST; d < info.day(); ++d){
                        certainPlayerSpecies.assign(info.executed(d), info.inquested(d));
                    }
                }else if(r == Role::WEREWOLF){ // 人狼の場合には味方がわかる
                   iterate(info.rolePlayers(Role::WEREWOLF), [&](int p){
                       certainPlayerSpecies.assign(p, Species::WEREWOLF);
                   });
                }
            }
        };
        
        struct ThinkInfo : public Info{
            
            constexpr static double ATT_RATE_ESTIMATE = 0.6; // 推測発言の減衰率
            
            // <その日の発言情報>
            std::array<BitArray64<4, N_MAX_DAYS>, N_MAX_PLAYERS> anyTalkNum; // over, skip以外の会話回数(プレーヤーごと)
            
            // (表の)投票宣言
            std::array<BitArray64<4, 16>, N_MAX_DAYS> voteDec; // 最新の投票宣言
            std::array<BitArray64<4, 16>, N_MAX_DAYS> votedDecNum; // 合計得票数
            std::array<BitArray64<4, 16>, N_MAX_DAYS> votedDecNumRank; // 合計得票数多い順のプレーヤー番号
            
            // 裏の投票宣言
            std::array<BitArray64<4, 16>, N_MAX_DAYS> whisperVoteDec; // 最新の裏での処刑投票宣言
            std::array<BitArray64<4, 16>, N_MAX_DAYS> whisperVotedDecNum; // 合計得票数(狼の分はwhisper内優先)
            std::array<BitArray64<4, 16>, N_MAX_DAYS> whisperVotedDecNumRank; // 合計得票数多い順のプレーヤー番号
            
            // 襲撃宣言
            std::array<BitArray64<4, 16>, N_MAX_DAYS> attackVoteDec; // 最新の襲撃投票宣言
            std::array<BitArray64<4, 16>, N_MAX_DAYS> attackVotedDecNum; // 合計得票数
            
            // <継続的に記録していく情報>
            // カミングアウト
            BitArray64<4, 16> comingOutRole; // 最新
            BitArray64<4, 16> comingOutDate; // カミングアウトしたタイミング
            BitArray32<4, 8> comingOutedNum; // 各役職でカミングアウトした人数
            
            // 他の発言
            std::array<BitArray32<2, 16>, N_MAX_PLAYERS> divineResult; // 占い結果
            std::array<BitArray32<2, 16>, N_MAX_PLAYERS> inquestResult; // 霊媒結果
            
            // 推測発言の強度(0中心)
            std::array<std::array<std::array<char, N_ROLES>, N_MAX_PLAYERS>, N_MAX_PLAYERS> estimateDegree;
            
            std::array<BitArray64<4, 16>, 2> divinedNum; // 占われた数
            std::array<BitArray64<4, 16>, 2> inquestedNum; // 霊媒された数
            
            // <発言-行動の関係>
            BitArray64<4, 16> voteDecDayNum; // 投票宣言を最低1回行った日数
            BitArray64<4, 16> voteEqualDecNum; // 投票宣言通りに投票した日数
            BitArray64<4, 16> whisperVoteDecDayNum; // 裏での処刑投票宣言を最低1回行った日数
            BitArray64<4, 16> whisperVoteEqualDecNum; // 裏での処刑投票宣言通りに投票した日数
            BitArray64<4, 16> attackVoteDecDayNum; // 裏での襲撃投票宣言を最低1回行った日数
            BitArray64<4, 16> attackVoteEqualDecNum; // 裏での襲撃投票宣言通りに投票した日数
            
            // <各プレーヤーの主観的情報>
            std::array<SubjectiveInfo, N_MAX_PLAYERS> sub;
            
            void clearSubjectiveInfo()noexcept{ // 各プレーヤーの主観的情報を更新する
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    sub[p].clear();
                }
            }
            
            void clearConversation()noexcept{ // 会話系情報をクリア
                
                voteDec.fill(0);
                whisperVoteDec.fill(0);
                attackVoteDec.fill(0);
                
                votedDecNum.fill(0);
                whisperVotedDecNum.fill(0);
                attackVotedDecNum.fill(0);
                
                voteDecDayNum.clear(); voteEqualDecNum.clear();
                whisperVoteDecDayNum.clear(); whisperVoteEqualDecNum.clear();
                attackVoteDecDayNum.clear(); attackVoteEqualDecNum.clear();
                
                comingOutRole.clear(); comingOutDate.clear();
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    comingOutRole.set(p, NONE_ROLE);
                }
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    divineResult[p].fill(NONE_SPECIES);
                    inquestResult[p].fill(NONE_SPECIES);
                }
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    for(int t = 0; t < N_MAX_PLAYERS; ++t){
                        estimateDegree[p][t].fill(0);
                    }
                }
            }
            
            // その日限りの情報
            void pushVoteDec(int d, int p, int t){ // p が t に投票すると宣言した場合
                voteDec[d].assign(p, t);
            }
            void pushWhisperVoteDec(int d, int p, int t){ // 裏で p が t に投票すると宣言した場合
                whisperVoteDec[d].assign(p, t);
            }
            void pushAttackDec(int d, int p, int t){ // p が t を襲撃すると宣言した場合
                attackVoteDec[d].assign(p, t);
            }
            void pushComingOut(int d, int p, int r){ // d日目にpがrであるとCOした場合
                comingOutRole.assign(p, r);
                comingOutDate.assign(p, d);
            }
            bool anyComingOut(int p)const{ // pが少なくとも1回COしているかどうか
                return (comingOutRole[p] != NONE_ROLE);
            }
            
            void sumUpConversation(){ // 発言の結果を集計
                // カミングアウト状況をまとめる
                comingOutedNum.clear();
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    if(comingOutRole[p] != NONE_ROLE){
                        comingOutedNum.add(comingOutRole[p], 1);
                    }
                }
                
                // 占いと霊媒の状況をまとめる
                // 過去の報告と食い違った場合には最新の報告の方を信頼するため、数え直す
                divinedNum.fill(0); inquestedNum.fill(0);
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    for(int t = 0; t < N_MAX_PLAYERS; ++t){
                        if(p != t){ // 自分自身の結果報告は無視する
                            if(divineResult[p][t] != NONE_SPECIES){
                                divinedNum[divineResult[p][t]].add(t, 1);
                            }
                            if(inquestResult[p][t] != NONE_SPECIES){
                                inquestedNum[inquestResult[p][t]].add(t, 1);
                            }
                        }
                    }
                }
            }
            
            void addContinualConversation(int d){
                // 継続的な情報の当日分を加算する
                if(d > 0){
                    // 推測発言の信憑性を減衰
                    for(int p = 0; p < N_MAX_PLAYERS; ++p){
                        for(int t = 0; t < N_MAX_PLAYERS; ++t){
                            for(int r = 0; r < N_ROLES; ++r){
                                int deg = estimateDegree[p][t][r];
                                deg /= 2; // TODO : 減衰率
                                estimateDegree[p][t][r] = deg;
                            }
                        }
                    }
                }
                
                for(int i = 0; i < NTalks(d); ++i){
                    const Talk t = talk(d, i);
                    int p = t.player();
                    if(t.any()){
                        anyTalkNum[p].add(d, 1); // 何からの発言の回数
                    }
                    switch(t.type()){
                        case OVER: case SKIP: break;
                        case ESTIMATE: { // 推定
                            estimateDegree[p][t.dst()][t.role()] = +127;
                        }break;
                        case COMINGOUT: { // CO
                            pushComingOut(d, p, t.role());
                        }break;
                        case DIVINED: { // 占い結果
                            if(comingOutRole[p] != Role::SEER){ // COしていない場合はCO宣言とみなす
                                pushComingOut(d, p, Role::SEER);
                            }
                            divineResult[p].assign(t.dst(), t.role());
                        }break;
                        case INQUESTED: { // 霊媒結果
                            if(comingOutRole[p] != Role::MEDIUM){ // COしていない場合はCO宣言とみなす
                                pushComingOut(d, p, Role::MEDIUM);
                            }
                            inquestResult[p].assign(t.dst(), t.role());
                        }break;
                        case GUARDED: { // 護衛結果
                            if(comingOutRole[p] != Role::BODYGUARD){ // COしていない場合はCO宣言とみなす
                                pushComingOut(d, p, Role::BODYGUARD);
                            }
                        }break;
                        case AGREE: {
                            auto agr = searchAgreeChain<1>(t.day(), t.category(), t.index());
                            Statement stm = agr.first;
                            if(stm.type() == ESTIMATE){
                                estimateDegree[p][stm.dst()][stm.role()] = +127;
                            }
                        }break;
                        case DISAGREE: {
                            auto agr = searchAgreeChain<0>(t.day(), t.category(), t.index());
                            Statement stm = agr.first;
                            if(stm.type() == ESTIMATE){
                                estimateDegree[p][stm.dst()][stm.role()] = -127;
                            }
                        }break;
                        default: break;
                    }
                }
            }
            
            void setDailyConversation(int d){
                // 当日の情報のみまとめる（人狼や客観的情報者なら全て知っている）
                voteDec[d].clear();
                whisperVoteDec[d].clear();
                attackVoteDec[d].clear();
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    voteDec[d].set(p, p); // 自分自身を投票としておく
                    whisperVoteDec[d].set(p, p); // 自分自身を投票としておく
                    attackVoteDec[d].set(p, p); // 自分自身を投票としておく
                }
                for(int i = 0; i < NTalks(d); ++i){
                    Talk t = talk(d, i);
                    if(t.type() == VOTE){
                        pushVoteDec(d, t.player(), t.dst());
                    }
                }
                for(int i = 0; i < NWhispers(d); ++i){
                    Whisper w = whisper(d, i);
                    if(w.type() == VOTE){
                        pushWhisperVoteDec(d, w.player(), w.dst());
                    }else if(w.type() == ATTACK){
                        pushAttackDec(d, w.player(), w.dst());
                    }
                }
                // 被投票数を数える
                votedDecNum[d].clear();
                whisperVotedDecNum[d].clear();
                attackVotedDecNum[d].clear();
                
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    // 表の投票宣言状況
                    if(voteDec[d][p] != p){ // 有効な投票宣言
                        votedDecNum[d].add(voteDec[d][p], 1);
                    }
                }
                if(NWhispers(d) > 0){ // whisper情報も持っている場合
                    for(int p = 0; p < N_MAX_PLAYERS; ++p){
                        if(playerRole(p) == Role::WEREWOLF){ // 裏の投票宣言を信頼
                            if(whisperVoteDec[d][p] != p){ // 有効な投票宣言
                                whisperVotedDecNum[d].add(whisperVoteDec[d][p], 1);
                            }
                        }else{ // 表の投票宣言を使用
                            if(voteDec[d][p] != p){ // 有効な投票宣言
                               whisperVotedDecNum[d].add(voteDec[d][p], 1);
                            }
                        }
                        // 襲撃宣言状況
                        if(attackVoteDec[d][p] != p){ // 有効な襲撃宣言
                            attackVotedDecNum[d].add(attackVoteDec[d][p], 1);
                        }
                    }
                }
            }
            
            void addConversationActionRelation(int d){
                // 会話と行動の関係について1日分追記
                // 表の投票宣言通りに投票したか
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    if(p != voteDec[d][p]){ // 有効な投票宣言をした
                        voteDecDayNum.add(p, 1);
                        if(vote(d, p) == voteDec[d][p]){ // 宣言通りに投票した
                            voteEqualDecNum.add(p, 1);
                        }
                    }
                }
                if(NWhispers(d) > 0){
                    // 裏の投票宣言通りに投票したか
                    iterate(rolePlayers(Role::WEREWOLF), [&](int p){
                        if(p != whisperVoteDec[d][p]){ // 有効な投票宣言をした
                            whisperVoteDecDayNum.add(p, 1);
                            if(vote(d, p) == whisperVoteDec[d][p]){ // 宣言通りに投票した
                                whisperVoteEqualDecNum.add(p, 1);
                            }
                        }
                    });
                    // 裏の襲撃宣言通りに投票したか
                    iterate(rolePlayers(Role::WEREWOLF), [&](int p){
                        if(p != attackVoteDec[d][p]){ // 有効な襲撃宣言をした
                            attackVoteDecDayNum.add(p, 1);
                            if(attackVote(d, p) == attackVoteDec[d][p]){ // 宣言通りに投票した
                                attackVoteEqualDecNum.add(p, 1);
                            }
                        }
                    });
                }
            }
        
            void proceedConversation(int d){
                // 前日までの情報はすでに設定されていて、そこに当日の情報を追加
                // その日限りの情報もまとめる
                addContinualConversation(d);
                setDailyConversation(d);
                sumUpConversation();
            }
            void setFullConversation(int day){
                // d日目までのの全日での情報、d日目の当日の情報を一括でまとめる
                clearConversation();
                for(int d = 0; d <= day; ++d){
                    addContinualConversation(d);
                    setDailyConversation(d);
                }
                sumUpConversation();
            }
            
            template<class sample_t>
            void setSampleRoles(const sample_t& sample){
                // 役職配置サンプルにそって情報を設定
                // 全体情報を完全情報にする
                playerRole_ = sample;
                for(int r = 0; r < N_ROLES; ++r){
                    rolePlayers_[r].reset();
                }
                for(int p = 0; p < rule().players; ++p){
                    rolePlayers_[sample[p]].set(p);
                }
                /*for(int d = DAY_FIRST; d <= day(); ++d){
                    proceedDay(d);
                }*/
                for(int p = 0; p < rule().players; ++p){
                    sub[p].setRole(p, sample[p], *this);
                }
            }
            
            std::string toSurveyString(int d)const{
                std::ostringstream oss;
                oss << "e vote dec = " << voteDec[d] << votedDecNum[d] << endl;
                oss << "e wvote dec = " << whisperVoteDec[d] << whisperVotedDecNum[d] << endl;
                oss << "a vote dec = " << attackVoteDec[d] << attackVotedDecNum[d] << endl;
                oss << "CO = " << comingOutRole[d] << comingOutedNum[d] << endl;
                oss << "divineReport = " << endl;
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    oss << divineResult[p] << endl;
                }
                oss << "divinedNum = " << divinedNum[0] << divinedNum[1] << endl;
                oss << "inquestReport = " << endl;
                for(int p = 0; p < N_MAX_PLAYERS; ++p){
                    oss << inquestResult[p] << endl;
                }
                oss << "inquestedNum = " << inquestedNum[0] << inquestedNum[1] << endl;
                
                return oss.str();
            }
            
            ThinkInfo(const Info& arg) : Info(arg){}
        };
    }
}

#endif // AIWOLF_THOUSAND_THINKINFO_HPP_