/*
defines.h
Katsuki Ohto
*/

#ifndef _DEFINES_H_
#define _DEFINES_H_

// ゲーム以前の定義とか
// 時間計測もすぐに使えるようにここに入れておく
// 設定ファイルの内容を使うので、
// 設定ファイルはこれより先にインクルードする
// ただし設定ファイルのインクルード無しでもコンパイルエラーにならないようにする

// インクルード
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cassert>
#include <vector>
#include <string>
#include <algorithm>
#include <sys/time.h>

#if defined (HAVE_BMI2)
#include <immintrin.h>
#endif

#if defined (HAVE_SSE4)
#include <smmintrin.h>
#elif defined (HAVE_SSE2)
#include <emmintrin.h>
#endif

#if !defined(NDEBUG)
#define UNREACHABLE assert(0)
#elif defined(_MSC_VER) && !defined(__INTEL_COMPILER)
#define UNREACHABLE __assume(0)
#elif defined(__INTEL_COMPILER)
#define UNREACHABLE __assume(0)
#elif defined(__GNUC__) && (4 < __GNUC__ || (__GNUC__ == 4 && 4 < __GNUC_MINOR__))
#define UNREACHABLE __builtin_unreachable()
#else
#define UNREACHABLE assert(0)
#endif

#if defined(_MSC_VER) && !defined(__INTEL_COMPILER)
#define FORCE_INLINE __forceinline
#elif defined(__INTEL_COMPILER)
#define FORCE_INLINE inline
#elif defined(__GNUC__)
#define FORCE_INLINE __attribute__((always_inline)) inline
#else
#define FORCE_INLINE inline
#endif

// Unroller
template <int N> struct Unroller {
    template <typename T> FORCE_INLINE void operator () (T t) {
        Unroller<N-1>()(t);
        t(N-1);
    }
};
template <> struct Unroller<0> {
    template <typename T> FORCE_INLINE void operator () (T t) {}
};

// 整数定義
typedef long long int int64;
typedef unsigned long long int uint64;
typedef int int32;
typedef unsigned int uint32;
typedef short int16;
typedef unsigned short uint16;
typedef signed char int8;
typedef unsigned char uint8;

typedef unsigned long ulong;


// クロック計測関数(64ビットで値を返す)
static uint64 cputime(){
    unsigned int ax, dx;
    asm volatile("rdtsc\nmovl %%eax,%0\nmovl %%edx,%1":"=g"(ax),"=g"(dx): :"eax","edx");
    return ((uint64)(dx) << 32) + (uint64)(ax);
}

// タイムスタンプ読み込みにかかる時間
// 環境依存しそうなので0にしておく
constexpr uint64 CLOCK_LATENCY = 0ULL;

// デバッグ用

uint64 CLOCK_START;

static void tick()noexcept{
    CLOCK_START = cputime();
}

static void tock(){
    uint64 CLOCK_END = cputime();
    printf("clock=%lld\n", CLOCK_END - CLOCK_START - CLOCK_LATENCY);
}

timeval CLOCKMICS_START;

static void tickMicS()noexcept{
    gettimeofday(&CLOCKMICS_START, NULL);
}

static void tockMicS(){
    timeval CLOCKMICS_END;
    gettimeofday(&CLOCKMICS_END, NULL);
    long t = (CLOCKMICS_END.tv_sec - CLOCKMICS_START.tv_sec) * 1000000 + CLOCKMICS_END.tv_usec - CLOCKMICS_START.tv_usec;
    printf("%ld micsec.\n", t);
}

//腕時計クラス


class Clock{
private:
    uint64 c_start;
public:
    void start()noexcept{ c_start = cputime(); }
    uint64 stop()const noexcept{ return cputime() - c_start - CLOCK_LATENCY; }
    uint64 restart()noexcept{ // 結果を返し、0から再スタート
        uint64 tmp = cputime();
        uint64 old = c_start;
        c_start = tmp;
        return tmp - old - CLOCK_LATENCY;
    }
    
    constexpr Clock(): c_start(){}
};

class ClockMicS{
    //microsec単位
private:
    timeval t_start;
public:
    void start()noexcept{ gettimeofday(&t_start, NULL); }
    long stop()const noexcept{
        timeval t_end;
        gettimeofday(&t_end, NULL);
        long t = (t_end.tv_sec - t_start.tv_sec) * 1000000 + t_end.tv_usec - t_start.tv_usec;
        return t;
    }
    long restart()noexcept{ // 結果を返し、0から再スタート
        timeval t_end;
        gettimeofday(&t_end,NULL);
        long t = (t_end.tv_sec - t_start.tv_sec) * 1000000 + t_end.tv_usec - t_start.tv_usec;
        t_start = t_end;
        return t;
    }
    ClockMicS(){}
    ClockMicS(int m){ start(); }
};

class ClockMS{
    //millisec単位
private:
    timeval t_start;
public:
    void start()noexcept{ gettimeofday(&t_start, NULL); }
    long stop()const noexcept{
        timeval t_end;
        gettimeofday(&t_end, NULL);
        long t = (t_end.tv_sec - t_start.tv_sec) * 1000 + (t_end.tv_usec - t_start.tv_usec) / 1000;
        return t;
    }
    long restart()noexcept{ // 結果を返し、0から再スタート
        timeval t_end;
        gettimeofday(&t_end, NULL);
        long t = (t_end.tv_sec - t_start.tv_sec) * 1000 + (t_end.tv_usec - t_start.tv_usec) / 1000;
        t_start = t_end;
        return t;
    }
    ClockMS(){}
    ClockMS(int m){ start(); }
};

//出力
struct Space{
    int n;
    constexpr Space():n(0){}
    constexpr Space(const int arg):n(arg){}
};

static std::ostream& operator<<(std::ostream& os, const Space& sp){
    for(int i = 0; i < sp.n; ++i){ os << ' '; }
    return os;
}

template<typename T = int>
struct OutXY{
    T x, y;
    constexpr OutXY(T ax, T ay):x(ax), y(ay){}
};

template<typename T>
static std::ostream& operator<<(std::ostream& os, const OutXY<T>& arg){
    os << '(' << arg.x << ", " << arg.y << ')';
    return os;
}

//3値定数
constexpr int _YES = 1;
constexpr int _NO = 0;
constexpr int _BOTH = 2;

//constexprのmin,max
template<typename T>
constexpr T cmax(const T& a,const T& b){
    return a<b?b:a;
}

template<typename T>
constexpr T cmin(const T& a,const T& b){
    return a>b?b:a;
}

//出力等可能なアサーション

// アサーション用関数
void reportError( const char *functionName,
                 const char *fileName, int lineNumber,
                 std::string message );

// 条件x、命令等y
#ifdef NDEBUG
#define ASSERT(X, Y)
#else
#define ASSERT(X, Y)  if(!(X)){ Y; assert(X); assert(0); }
#endif // NDEBUG

// 浮動小数点がまともな値を取っているかどうかのアサーション
#define FASSERT(f, o) ASSERT(!std::isinf(f) && !std::isnan(f), cerr << (f) << endl;{o});

//標準ライブラリ使用
using std::cout;
using std::cerr;
using std::setw;
using std::endl;
using std::ostream;
using std::stringstream;
using std::max;
using std::min;

//出力
#ifdef DEBUG
#define DOUT std::cout
#define DERR std::cerr
#define DWAIT getchar()
#else
#define DERR 0 && std::cerr
#define DOUT 0 && std::cout
#define DWAIT
#endif

#ifndef MINIMUM
#define COUT cout
#define CERR cerr
#define CWAIT getchar()
#else
#define CERR 0 && cerr
#define COUT 0 && cout
#define CWAIT
#endif


//デバッグ用のブレークポイント
#define BREAKPOINT {volatile int a = 1; std::random_device seed_gen; std::mt19937 engine(seed_gen()); if(engine() % 2){ std::cerr << "break" << endl; }}

// 文字列処理
bool isPrefix(const std::string& s, const std::string& prefix){
    if (s.size() < prefix.size())
        return false;
    
    return s.substr(0, prefix.size()) == prefix;
}

bool isSuffix(const std::string& s, const std::string& suffix){
    if (s.size() < suffix.size())
        return false;
    
    return s.substr(s.size() - suffix.size()) == suffix;
}

bool contains(const std::string& s, const std::string& t){
    return s.find(t) != std::string::npos;
}

std::string trim(const std::string& s, const char dlm = ' '){
    std::string::size_type p1 = s.find_first_not_of(dlm);
    if (p1 == std::string::npos)
        return std::string();
    
    std::string::size_type p2 = s.find_last_not_of(dlm);
    return s.substr(p1, p2 - p1 + 1);
}

std::vector<std::string> split(const std::string& s, char separator){
    std::vector<std::string> result;
    std::string::size_type p = 0;
    std::string::size_type q;
    while ((q = s.find(separator, p)) != std::string::npos) {
        result.emplace_back(s, p, q - p);
        p = q + 1;
    }
    
    result.emplace_back(s, p);
    return result;
}

std::string join(const std::vector<std::string>& strs, const std::string& sep){
    if (strs.size() == 0)
        return std::string();
    
    std::ostringstream os;
    os << strs[0];
    for (size_t i = 1; i < strs.size(); ++i) {
        os << sep << strs[i];
    }
    return os.str();
}


//環境依存にはまらないように色々

#ifndef M_PI
#define M_PI 3.14159265358979343846
#endif

//#ifdef (_MSC_VER)
//#define constexpr
//#endif


#endif  // _DEFINES_H_