/*
 bitSet.hpp
 Katsuki Ohto
 */

#ifndef UTIL_BITSET_HPP_
#define UTIL_BITSET_HPP_

// ビットセットを定義
// サイズ固定
// std::bitsetより色々機能が多い

#include <cassert>
#include <string>
#include <iostream>
#include <sstream>

#include "../defines.h"
//#include "type.hpp"
#include "bitOperation.hpp"
#include "container.hpp"

#include "bitArray.hpp"

template<typename data_t, int MAX_SIZE_ = (sizeof(data_t) * 8), typename atomic_data_t = data_t>
class BitSetInRegister{
    
public:
    using this_type_t = BitSetInRegister<data_t, MAX_SIZE_, atomic_data_t>;
    using size_t = std::size_t; // unsigned int;
    using entry_t = bool;
    
    constexpr BitSetInRegister() :dat_(){}
    constexpr BitSetInRegister(const data_t arg) : dat_(arg){}
    
    // operator
    constexpr operator data_t()const noexcept{ return static_cast<data_t>(dat_); }
    
    BitSetInRegister& operator=(const data_t arg)noexcept{ dat_ = arg; return *this; }
    BitSetInRegister& operator+=(const data_t arg)noexcept{ dat_ += arg; return *this; }
    BitSetInRegister& operator-=(const data_t arg)noexcept{ dat_ -= arg; return *this; }
    BitSetInRegister& operator*=(const data_t arg)noexcept{ dat_ += arg; return *this; }
    BitSetInRegister& operator/=(const data_t arg)noexcept{ dat_ -= arg; return *this; }
    BitSetInRegister& operator|=(const data_t arg)noexcept{ dat_ |= arg; return *this; }
    BitSetInRegister& operator&=(const data_t arg)noexcept{ dat_ &= arg; return *this; }
    BitSetInRegister& operator^=(const data_t arg)noexcept{ dat_ ^= arg; return *this; }
    BitSetInRegister& operator<<=(const size_t arg)noexcept{ dat_ <<= arg; return *this; }
    BitSetInRegister& operator>>=(const size_t arg)noexcept{ dat_ >>= arg; return *this; }
        
    static void assert_index(const size_t p){
        ASSERT(0 <= p && p < MAX_SIZE_, std::cerr << p << " in " << MAX_SIZE_ << std::endl; );
    }
        
    constexpr static data_t mask(const size_t p)noexcept{ return (MASK_BIT_ << p); }
    constexpr static data_t lower_mask(const size_t p)noexcept{ return (MASK_BIT_ << p) - static_cast<data_t>(1); }
    constexpr static data_t mask(const size_t p0, const size_t p1)noexcept{
        return lower_mask(p1 + 1) & (~lower_mask(p0));
    }
    constexpr static data_t full_mask()noexcept{ return static_cast<data_t>(-1); }
        
    void set(const size_t p)noexcept{
        assert_index(p);
        dat_ |= mask(p);
    }
    void flip(const size_t p)noexcept{
        assert_index(p);
        dat_ ^= mask(p);
    }
    void reset(const size_t p)noexcept{
        assert_index(p);
        dat_ &= ~mask(p);
    }
        
    void push0(const size_t p)noexcept{
        assert_index(p);
        dat_ = (dat_&lower_mask(p)) | ((dat_&(~lower_mask(p))) << 1);
    }
    void push1(const size_t p)noexcept{
        assert_index(p);
        dat_ = (dat_&lower_mask(p)) | ((dat_&(~lower_mask(p))) << 1) | mask(p);
    }
        
    void remove(const size_t p)noexcept{
        assert_index(p);
        dat_ = (dat_ & lower_mask(p)) | ((dat_ >> 1) & (~lower_mask(p)));
    }
    //void remove_and_fill_by_msb(const size_t p)noexcept{
    //    assert_index(p);
    //    dat_ = (dat_ & lower_mask(p)) | ((static_cast<signed_type<data_t>::type>(dat_) >> 1) & (~lower_mask(p)));
    //}

    void reset()noexcept{ dat_ = static_cast<data_t>(0); }
    void flip()noexcept{ dat_ = ~dat_; }

    void set(const size_t p0, const size_t p1)noexcept{
        dat_ |= mask(p0) | mask(p1);
    }
    void set(const size_t p0, const size_t p1, const size_t p2)noexcept{
        dat_ |= mask(p0) | mask(p1) | mask(p2);
    }
    void set(const size_t p0, const size_t p1, const size_t p2, const size_t p3)noexcept{
        dat_ |= mask(p0) | mask(p1) | mask(p2) | mask(p3);
    }
    void set(const size_t p0, const size_t p1, const size_t p2, const size_t p3, const size_t p4)noexcept{
        dat_ |= mask(p0) | mask(p1) | mask(p2) | mask(p3) | mask(p4);
    }
    void set(const size_t p0, const size_t p1, const size_t p2, const size_t p3, const size_t p4, const size_t p5)noexcept{
        dat_ |= mask(p0) | mask(p1) | mask(p2) | mask(p3) | mask(p4) | mask(p5);
    }

    void fill()noexcept{
        dat_ = full_mask();
    }
    void fill_through_back(const size_t p)noexcept{
        assert_index(p);
        dat_ |= ~lower_mask(p);
    }
    void fill(const size_t p0, const size_t p1)noexcept{
        assert_index(p0); assert_index(p1);
        dat_ |= lower_mask(p1 + 1) & (~lower_mask(p0));
    }

    constexpr data_t test(const size_t p)const noexcept{
        return dat_ & mask(p);
    }
    constexpr entry_t get(const size_t p)const noexcept{
        return (dat_ >> p) & MASK_BIT_;
    }

    template<int NN, typename other_atomic_data_t>
    constexpr bool holds(const BitSetInRegister<data_t, NN, other_atomic_data_t>& arg)const noexcept{
        return holdsBits<data_t>(dat_, arg.dat_);
    }

    constexpr data_t holds(const size_t p0)const noexcept{
        return test(p0);
    }
    constexpr bool holds(const size_t p0, const size_t p1)const noexcept{
        return holdsBits<data_t>(dat_, mask(p0) | mask(p1));
    }
    constexpr bool holds(const size_t p0, const size_t p1, const size_t p2)const noexcept{
        return holdsBits<data_t>(dat_, mask(p0) | mask(p1) | mask(p2));
    }
    constexpr bool holds(const size_t p0, const size_t p1, const size_t p2, const size_t p3)const noexcept{
        return holdsBits<data_t>(dat_, mask(p0) | mask(p1) | mask(p2) | mask(p3));
    }
    constexpr bool holds(const size_t p0, const size_t p1, const size_t p2, const size_t p3, const size_t p4)const noexcept{
        return holdsBits<data_t>(dat_, mask(p0) | mask(p1) | mask(p2) | mask(p3) | mask(p4));
    }
    constexpr bool holds(const size_t p0, const size_t p1, const size_t p2, const size_t p3, const size_t p4, const size_t p5)const noexcept{
        return holdsBits<data_t>(dat_, mask(p0) | mask(p1) | mask(p2) | mask(p3) | mask(p4) | mask(p5));
    }

    constexpr bool is_only(const size_t p)const noexcept{
        return (dat_ == (static_cast<data_t>(1)<<p));
    }
    
    constexpr data_t anyExcept(const size_t p)const noexcept{ return dat_ & (~mask(p));}
    
    constexpr data_t any()const noexcept{ return dat_; }
    size_t count()const noexcept{ return countBits<data_t>(dat_); }
    size_t bsf()const noexcept{ return ::bsf<data_t>(dat_); }
    size_t bsr()const noexcept{ return ::bsr<data_t>(dat_); }

    void pop_lsb()noexcept{ dat_ &= (dat_ - static_cast<data_t>(1)); }

    void init()noexcept{ dat_ = static_cast<data_t>(0); }

    constexpr data_t data()const noexcept{ return dat_; }
        
    // 配列化。ビット数やデータ数がオーバーしたときの動作は未定義
    template<int N, int SIZE = 64 / N>
    BitArray64<N, SIZE> toArray64()const{
        BitArray64<N, SIZE> ret(0);
        auto bs = *this;
        while (bs.any()){
            ret.insert(0, bs.bsf());
            bs.pop_lsb();
        }
        return ret;
    }
    template<int N, int SIZE = 32 / N>
    BitArray32<N, SIZE> toArray32()const{
        BitArray32<N, SIZE> ret(0);
        auto bs = *this;
        while (bs.any()){
            ret.insert(0, bs.bsf());
            bs.pop_lsb();
        }
        return ret;
    }
        
        
    std::string toSetString()const{
        std::ostringstream oss;
        this_type_t tmp = data();
        oss << "{";
        while(tmp.any()){
            size_t i = tmp.bsf();
            oss << i;
            tmp.pop_lsb();
            if(tmp.any()){
                oss << ",";
            }
        }
        oss << "}";
        return oss.str();
    }

private:
    constexpr static data_t MASK_BIT_ = static_cast<data_t>(1);
    
    atomic_data_t dat_;
};

template<typename data_t, int MAX_SIZE>
std::ostream& operator<<(std::ostream& out, const BitSetInRegister<data_t, MAX_SIZE>& arg){ // output
    out << "[";
    for (int i = 0; i < MAX_SIZE; ++i){
        out << int(arg.get(i));
    }
    out << "]";
    return out;
}

template<typename data_t, int MAX_SIZE, typename callback_t>
void iterate(BitSetInRegister<data_t, MAX_SIZE> bs, const callback_t& callback)noexcept{
    while (bs.any()){
        callback(bs.bsf());
        bs.pop_lsb();
    }
}
    template<typename data_t, int MAX_SIZE, typename callback_t>
    void iterateWithIndex(BitSetInRegister<data_t, MAX_SIZE> bs, const callback_t& callback)noexcept{
        int cnt = 0;
        while (bs.any()){
            callback(cnt, bs.bsf());
            bs.pop_lsb();
            ++cnt;
        }
    }

template<typename data_t, int MAX_SIZE, typename callback_t>
void iterateWithoutFirstCheck(BitSetInRegister<data_t, MAX_SIZE> bs, const callback_t& callback)noexcept{
    do{
        callback(bs.bsf());
        bs.pop_lsb();
    } while (bs.any());
}

template<typename data_t, int MAX_SIZE, typename callback_t>
void iterateExcept(BitSetInRegister<data_t,MAX_SIZE> bs, const std::size_t p, const callback_t& callback)noexcept{
    BitSetInRegister<data_t, MAX_SIZE>::assert_index(p);
    bs.reset(p);
    while (bs.any()){
        callback(bs.bsf());
        bs.pop_lsb();
    }
}

template<typename data_t, int MAX_SIZE, typename callback_t>
void iterateWithoutFirstCheckExcept(BitSetInRegister<data_t, MAX_SIZE> bs, const std::size_t p, const callback_t& callback)noexcept{
    BitSetInRegister<data_t, MAX_SIZE>::assert_index(p);
    bs.reset(p);
    do{
        callback(bs.bsf());
        bs.pop_lsb();
    } while (bs.any());
}

template<typename data_t, int MAX_SIZE, typename callback_t>
int search(BitSetInRegister<data_t, MAX_SIZE> bs, const callback_t& callback)noexcept{
    while(bs.any()){
        const std::size_t idx = bs.bsf();
        if (callback(idx)){ return idx; }
        bs.pop_lsb();
    }
    return -1;
}
using BitSet8 = BitSetInRegister<uint8>;
using BitSet16 = BitSetInRegister<uint16>;
using BitSet32 = BitSetInRegister<uint32>;
using BitSet64 = BitSetInRegister<uint64>;

#endif // UTIL_BITSET_HPP_