/*
 random.hpp
 Katsuki Ohto
 */

#ifndef UTIL_RANDOM_HPP_
#define UTIL_RANDOM_HPP_

//乱数生成

std::random_device seed_gen;
std::default_random_engine engine(seed_gen());
std::uniform_real_distribution<> dist(0.0, 1.0);

//指数分布
struct Exp{
    double lambda;
    double rand()const{
        return (-1.0 / lambda * (1.0 - dist(engine)));
    }
    explicit Exp(double l){ lambda = l; }
};

//ガンマ分布
struct Gamma{
    double a;
    double rand()const{
        double x, y, z;
        double u, v, w, b, c, e;
        int accept = 0;
        if (a < 1)
        {
            Exp ex(1);
            e = ex.rand();
            do {
                x = pow(dist(engine), 1 / a);
                y = pow(dist(engine), 1 / (1 - a));
            } while (x + y > 1);
            return (e * x / (x + y));
        } else {
            b = a - 1;
            c = 3 * a - 0.75;
            do {
                u = dist(engine);
                v = dist(engine);
                w = u * (1 - u);
                y = sqrt(c / w) * (u - 0.5);
                x = b + y;
                if (x >= 0)
                {
                    z = 64 * w * w * w * v * v;
                    if (z <= 1 - (2 * y * y) / x)
                    {
                        accept = 1;
                    } else {
                        if (log(z) < 2 * (b * log(x / b) - y))
                            accept = 1;
                    }
                }
            } while (accept != 1);
            return x;
        }
    }
    void add(double arg)noexcept{
        a += arg;
    }
    void subtr(double arg)noexcept{
        a -= arg;
    }
    void set(double arg)noexcept{
        a = arg;
    }
};

//ベータ分布
struct Beta{
    Gamma gam0,gam1;
    double rand()const{
        double r1 = gam0.rand();
        double r2 = gam1.rand();
        return r1 / (r1 + r2);
    }
    double size()const noexcept{
        return gam0.a + gam1.a;
    }
    double mean()const{
        return gam0.a / (gam0.a + gam1.a);
    }
    double variance()const{
        double i = gam0.a;
        double j = gam1.a;
        double sum = i + j;
        
        assert(sum * sum * (sum + 1.0));
        
        return (i * j) / (sum * sum * (sum + 1.0));
    }
    double sd()const{
        return std::sqrt(variance());
    }
    void add(const Beta& arg)noexcept{
        gam0.add(arg.gam0.a);
        gam1.add(arg.gam1.a);
    }
    void subtr(const Beta& arg)noexcept{
        gam0.subtr(arg.gam0.a);
        gam1.subtr(arg.gam1.a);
    }
    void mul(const double m)noexcept{
        gam0.set(gam0.a * m);
        gam1.set(gam1.a * m);
    }
    void rev()noexcept{
        std::swap(gam0, gam1);
    }
    
    void set(const double argA, const double argB)noexcept{
        gam0.set(argA);
        gam1.set(argB);
    }
    
    void resize(double h){
        //サイズをhにする
        double s=size();
        
        assert( s );
        
        double h_s=h / s;
        gam0.set( gam0.a * h_s );
        gam1.set( gam1.a * h_s );
    }
    
    bool exam()const noexcept{
        if(gam0.a < 0 || gam1.a < 0){ return false; }
        if(gam0.a == 0 && gam1.a == 0){ return false; }
        return true;
    }
    
    Beta(const double argA, const double argB){
        set(argA,argB);
    }
    Beta(){}
    ~Beta(){}
};

std::ostream& operator<<(std::ostream& out, const Beta& b){
    out << "( " << b.gam0.a << " , " << b.gam1.a << " )";
    return out;
}

//ディリクレ分布
//各確率変数が報酬を持つとし、その報酬を返す
template<int N>
class Dir{
public:
    Gamma gam[N];
    double *rew;
    double rand()const{
        double sum = 0.0;
        double sum_rew = 0.0;
        for(auto i = 0; i < N ; ++i){
            double r = gam[i].rand();
            sum += r;
            sum_rew += rew[i] * r;
        }
        sum_rew /= sum;
        return sum_rew;
    }
    Dir(double *argRew, const double argX[]){
        for(auto i = 0; i < N; ++i){
            gam[i].set(argX[i]);
        }
        rew = argRew;
    }
    Dir(){}
    void set(double *argRew, const double argX[]){
        for(auto i =  0; i < N; ++i){
            gam[i].set(argX[i]);
        }
        rew = argRew;
    }
};

//正規分布
template<typename f_t>
struct Norm{
    f_t mu;
    f_t sigma;
    
    f_t rand()const{
        //Box-Muller
        f_t r1 = dist(engine);
        f_t r2 = dist(engine);
        f_t z1 = std::sqrt(-2.0 * std::log(r1)) * std::cos(2.0 * M_PI * r2);
        return z1 * sigma + mu;
    }
    template<class dice_t>
    f_t rand(dice_t *const pdice)const{
        //Box-Muller
        f_t r1 = pdice->drand();
        f_t r2 = pdice->drand();
        f_t z1 = std::sqrt(-2.0 * std::log(r1)) * std::cos(2.0 * M_PI * r2);
        return z1 * sigma + mu;
    }
    template<class dice_t>
    void rand(f_t *const pa, f_t *const pb, dice_t *const pdice)const{
        //2つ同時に発生させる
        f_t r1 = pdice->drand();
        f_t r2 = pdice->drand();
        
        //CERR<<r1<<","<<r2<<" : ";
        f_t z1 = std::sqrt(-2.0 * std::log(r1)) * std::cos(2.0 * M_PI * r2);
        f_t z2 = std::sqrt(-2.0 * std::log(r1)) * std::sin(2.0 * M_PI * r2);
        *pa = z1 * sigma + mu;
        *pb = z2 * sigma + mu;
        return;
    }
    
    constexpr Norm():mu(),sigma(){}
    constexpr Norm(f_t argMu,f_t argSigma)
    :mu(argMu),sigma(argSigma){}
    
    void set(f_t argMu, f_t argSigma)noexcept{
        mu = argMu;
        sigma = argSigma;
    }
};

#endif // UTIL_RANDOM_HPP_